#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-04-08
# @Filename: json.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import asyncio
import json

from typing import Any, Callable, Coroutine, Dict, List, TypeVar

from ..command import Command


__all__ = ["JSONParser"]


T = TypeVar("T", bound=Command)


DEFAULT_CALLBACKS = {}


class JSONParser:
    """A parser that receives commands and arguments as a JSON string.

    See :ref:`json-parser` for details on implementation and use-cases.

    """

    #: list: Additional arguments to be passed to each command in the parser.
    #: Note that the command is always passed first.
    parser_args: List[Any] = []

    #: dict: Mapping of command verb to callback coroutine.
    callbacks: Dict[str, Callable[..., Coroutine]] = DEFAULT_CALLBACKS

    def parse_command(self, command: T) -> T:
        """Parses a user command.

        The command string must be a serialised JSON-like string that contains at
        least a keyword ``command`` with the name of the callback, and any number of
        additional arguments which will be passed to it.
        """

        # This will pass the command as the first argument for each command.
        # If self.parser_args is defined, those arguments will be passed next.
        parser_args = [command]
        parser_args += self.parser_args

        # Empty command. Just finish the command.
        if not command.body:
            command.done()
            return command

        command.set_status(command.status.RUNNING)

        try:
            payload = json.loads(command.body)
        except json.JSONDecodeError as err:
            return command.fail(
                error=f"Cannot deserialise command string {command.body!r}: {err}"
            )

        if "command" not in payload:
            return command.fail(
                error=f"Command {command.body!r} does not contain "
                "a 'command' parameter."
            )

        verb = payload.pop("command")
        if verb not in self.callbacks:
            return command.fail(error=f"Cannot find a callback for command {verb!r}.")
        elif not asyncio.iscoroutinefunction(self.callbacks[verb]):
            return command.fail(error=f"Callback {verb!r} is not a coroutine function.")

        try:
            asyncio.create_task(self.callbacks[verb](*parser_args, payload))
        except Exception as err:
            return command.fail(
                error="Errored scheduling callback coroutine for "
                f"command {verb!r}: {err}"
            )

        return command



"""
String parser for SDSS-3 command and reply protocols

Refer to https://trac.sdss3.org/wiki/Ops/Parsing for details. Unit and
timing tests, together with an alternative implementation using
pyparsing, are available in the SDSS-3 repository in
tops.sdss3.protocols.

Uses a copy of the PLY parser package
(http://www.dabeaz.com/ply/ply.html) that is included in this
distribution.
"""

# type: ignore

# Created 10-Oct-2008 by David Kirkby (dkirkby@uci.edu)

import re

from .messages import (
    ActorReplyHeader,
    Command,
    Keyword,
    MessageError,
    RawKeyword,
    Reply,
    ReplyHeader,
)
from .ply import lex, yacc


name_pattern = re.compile("[A-Za-z][A-Za-z0-9_.]*$")


class ParseError(Exception):
    pass


class TemporaryBase(object):
    """
    Serves as a temporary base class that will be substituted later
    """

    pass


class ReplyParser(TemporaryBase):
    """
    Specifies the parsing rules for reply strings

    The base class will eventually be ParserBase but is not declared
    above since PLY applies grammar rules in the order in which they are
    declared.
    """

    _name = "(?:[A-Za-z][A-Za-z0-9_]*)"
    _extra = r"(?:\.[A-Za-z][A-Za-z0-9_.]*)"
    _number = "0|[1-9][0-9]*"
    hdr_pattern = re.compile(
        r"(%s?)\.(%s)(%s?)[ \t]+(%s)[ \t]+(%s)[ \t]+(.)[ \t]+"
        % (_name, _name, _extra, _number, _name)
    )

    # lexical tokens
    tokens = ["EQUALS", "COMMA", "VALUE", "NAME_OR_VALUE", "QUOTED", "SEMICOLON"]

    def p_reply(self, p):
        "reply : keywords"
        p[0] = Reply(header=self.header, keywords=p[1], string=self.string)

    def p_empty_reply(self, p):
        "reply : "
        p[0] = Reply(header=self.header, keywords=[], string=self.string)

    def p_reply_keywords(self, p):
        "keywords : keywords SEMICOLON keyword"
        p[0] = p[1]
        p[0].append(p[3])

    def p_one_keyword(self, p):
        "keywords : keyword"
        p[0] = [p[1]]

    t_SEMICOLON = ";"

    def unwrap(self):
        matched = ReplyParser.hdr_pattern.match(self.string)
        if not matched:
            raise ParseError("Badly formed reply header in: %s" % self.string)
        self.header = ReplyHeader(*matched.groups())
        return self.string[matched.end() :]


class ActorReplyParser(ReplyParser):
    """
    Specifies the parsing rules for reply strings from actors

    The basic format is:
    commandID userID code data

    The base class will eventually be ParserBase but is not declared
    above since PLY applies grammar rules in the order in which they are
    declared.
    """

    hdr_pattern = re.compile(
        "(%s)[ \t]+(%s)[ \t]+(.)[ \t]+" % (ReplyParser._number, ReplyParser._number)
    )

    def unwrap(self):
        matched = ActorReplyParser.hdr_pattern.match(self.string)
        if not matched:
            raise ParseError("Badly formed reply header in: %s" % self.string)
        self.header = ActorReplyHeader(*matched.groups())
        return self.string[matched.end() :]


class CommandParser(TemporaryBase):
    """
    Specifies the parsing rules for command strings

    The base class will eventually be ParserBase but is not declared
    above since PLY applies grammar rules in the order in which they are
    declared.
    """

    # lexical tokens
    tokens = ["EQUALS", "COMMA", "VALUE", "NAME_OR_VALUE", "QUOTED", "RAW", "LINE"]

    def p_verb_with_values_command(self, p):
        "command : verb_with_values keywords"
        p[0] = Command(name=p[1][0], values=p[1][1:], keywords=p[2], string=self.string)

    def p_verb_no_values_command(self, p):
        "command : NAME_OR_VALUE NAME_OR_VALUE values keywords"
        p[0] = Command(name=p[1], keywords=[Keyword(p[2], p[3])], string=self.string)
        p[0].keywords.extend(p[4])

    def p_raw_command(self, p):
        "command : NAME_OR_VALUE RAW LINE"
        p[0] = Command(name=p[1], keywords=[RawKeyword(p[3])], string=self.string)

    def p_ambiguous_command(self, p):
        "command : NAME_OR_VALUE NAME_OR_VALUE keywords"
        p[0] = Command(name=p[1], keywords=[Keyword(p[2])], string=self.string)
        p[0].keywords.extend(p[3])

    def p_verb_only_command(self, p):
        "command : NAME_OR_VALUE"
        p[0] = Command(name=p[1], string=self.string)

    def p_verb_add_values(self, p):
        "verb_with_values : verb_with_values COMMA value"
        p[0] = p[1]
        p[0].append(p[3])

    def p_verb_with_two_values(self, p):
        "verb_with_values : NAME_OR_VALUE NAME_OR_VALUE COMMA NAME_OR_VALUE"
        p[0] = [p[1], p[2], p[4]]

    def p_verb_with_one_value(self, p):
        """verb_with_values : NAME_OR_VALUE VALUE
        | NAME_OR_VALUE QUOTED"""
        p[0] = [p[1], p[2]]

    def p_command_keywords(self, p):
        "keywords : keywords keyword"
        p[0] = p[1]
        p[0].append(p[2])

    def p_no_keywords(self, p):
        "keywords : "
        p[0] = []

    def p_raw_keyword(self, p):
        "keyword : RAW LINE"
        p[0] = RawKeyword(p[2])

    # lexical analysis states
    states = (("RAW", "exclusive"),)

    # capture inline whitespace in RAW mode
    t_RAW_ignore = ""

    def t_RAW_LINE(self, t):
        r"\s*=(?P<value>.*)"
        t.value = self.lexer.lexmatch.group("value")
        self.lexer.begin("INITIAL")
        return t

    def word_token(self, t):
        if t.value.lower() == "raw":
            t.type = "RAW"
            t.lexer.begin("RAW")
            return t
        else:
            return ParserBase.word_token(self, t)

    def unwrap(self):
        return self.string


class ParserBase(object):
    """
    Specifies the parsing rules common to all messages

    Although this is the base class for command and reply parsers we
    must define it after those classes since PLY applies grammar rules
    in the order in which they are declared.
    """

    def __init__(self, debug=0):
        """
        Initializes a new parser object

        A debug level of 1 will print out the lex rules on
        initialization and the yacc state in case of an error. Level 2
        will also provide a trace of the yacc analysis.
        """
        self.debug = debug
        self.lexer = lex.lex(object=self, debug=debug)
        self.engine = yacc.yacc(module=self, debug=debug, write_tables=0)

    def p_keyword_with_values(self, p):
        "keyword : NAME_OR_VALUE values"
        p[0] = Keyword(p[1], p[2])

    def p_bare_keyword(self, p):
        "keyword : NAME_OR_VALUE"
        p[0] = Keyword(p[1])

    def p_many_values(self, p):
        "values : values COMMA value"
        p[0] = p[1]
        p[0].append(p[3])

    def p_values(self, p):
        "values : EQUALS value"
        p[0] = [p[2]]

    def p_value(self, p):
        """value : NAME_OR_VALUE
        | VALUE
        | QUOTED"""
        p[0] = p[1]

    # ignore redundant inline whitespace
    t_ignore = " \t"

    # single-character literal tokens
    t_EQUALS = "="
    t_COMMA = ","

    def word_token(self, t):
        if name_pattern.match(t.value):
            t.type = "NAME_OR_VALUE"
        else:
            t.type = "VALUE"
        return t

    def t_WORD(self, t):
        r'[^"\'\s=,;]+'
        return self.word_token(t)

    def t_QUOTED1(self, t):
        r"'(?P<value>(?:[^\\'\n]|\\.)*?)'"
        t.type = "QUOTED"
        t.value = t.lexer.lexmatch.group("value")
        return t

    def t_QUOTED2(self, t):
        r'"(?P<value>(?:[^\\"\n]|\\.)*?)"'
        t.type = "QUOTED"
        t.value = t.lexer.lexmatch.group("value")
        return t

    def p_error(self, tok):
        """
        Handles parse errors
        """
        # reset the lexer in case we were doing some modal processing
        # (this should only happen for the raw command keyword)
        # see http://www.dabeaz.com/ply/ply.html#ply_nn21
        self.lexer.begin("INITIAL")
        if not tok:
            raise ParseError("Unexpected end of input")
        raise ParseError(
            "Unexpected %s parse token (%r) in:\n%s"
            % (tok.type, tok.value[: min(len(tok.value), 20)], self.string)
        )

    def t_ANY_error(self, tok):
        """
        Handles lexical analysis errors
        """
        if not tok:
            raise ParseError("Undefined lexical analysis error")
        raise ParseError(
            "Lexical analysis error at %r" % tok.value[: min(len(tok.value), 20)]
        )

    def tokenize(self, string):
        """
        Generates the lexical tokens found in the message string
        """
        string = string.rstrip("\n")
        self.lexer.input(string)
        tok = self.lexer.token()
        while tok:
            yield tok
            tok = self.lexer.token()

    def parse(self, string):
        """
        Returns a parsed representation of the message string

        Any trailing newline characters are removed before parsing.
        Raises an exception if the string cannot be parsed.
        """
        self.string = string.rstrip("\n")
        try:
            body = self.unwrap()
            return self.engine.parse(body, lexer=self.lexer, debug=self.debug)
        except MessageError as e:
            raise ParseError(str(e))


# Replace the parser base classes. This is necessary since PLY applies YACC rules
# in the order in which they are declared in the source file, but python expects
# that a base class is declared before any subclasses.

ReplyParser.__bases__ = (ParserBase,)
CommandParser.__bases__ = (ParserBase,)



"""
Generates HTML files via python declarations
"""

# type: ignore

# Created 28-Jun-2008 by David Kirkby (dkirkby@uci.edu)

from getpass import getuser
from html import escape
from os.path import splitext
from socket import gethostname
from sys import argv
from time import ctime


class HTMLDocumentError(Exception):
    """
    Represents an error that occurs while declaring an HTML document.
    """

    pass


class Head:
    """
    Declares the head element of an HTML document.
    """

    iconTypes = {
        ".ico": "image/vnd.microsoft.icon",
        ".gif": "image/gif",
        ".png": "image/png",
    }

    def __init__(self, title=None, base=None, icon=None, css=None, js=None, raw=None):
        self.type = type
        self.title = title
        self.base = base
        self.icon = icon
        self.css = []
        if css:
            self.css.append(css)
        self.js = []
        if js:
            self.js.append(js)
        self.raw = []
        if raw:
            self.raw.append(raw)

    def __str__(self):
        s = "<head>\n"
        s += '  <meta http-equiv="Content-Type" content="text/html; charset=utf-8">\n'
        s += (
            '  <meta name="description" '
            'content="Document created automatically on %s by %s@%s using %s">\n'
            % (ctime(), getuser(), gethostname(), argv[0])
        )
        if self.title:
            s += "  <title>%s</title>\n" % self.title
        if self.base and ("href" in self.base or "target" in self.base):
            s += "  <base"
            if "href" in self.base:
                s += ' href="%s"' % self.base["href"]
            else:
                s += ' href=""'
            # should check that target only uses [A-Za-z] or is one of the predefined
            # targets (http://htmlhelp.com/reference/html40/values.html#frametarget)
            if "target" in self.base:
                s += ' target="%s"' % self.base["target"]
            s += ">\n"
        if self.icon:
            iconExtension = splitext(self.icon)[1]
            if iconExtension not in Head.iconTypes:
                raise HTMLDocumentError(
                    "Unsupported icon extension: %s" % iconExtension
                )
            s += '  <link rel="icon" type="%s" href="%s" />\n' % (
                Head.iconTypes[iconExtension],
                self.icon,
            )
        for style in self.css:
            # is this an external reference? look for a one-liner
            if style.count("\n") == 0:
                s += '  <link rel="stylesheet" type="text/css" href="%s">\n' % style
            else:
                s += '  <style type="text/css">\n%s</style>\n' % style
        for script in self.js:
            # is this an external reference? look for a one-liner
            if script.count("\n") == 0:
                s += '  <script type="text/javascript" src="%s"></script>\n' % script
            else:
                s += '  <script type="text/javascript">\n%s</script>\n' % script
        for raw in self.raw:
            s += "  %s\n" % raw
        s += "</head>\n"
        return s


class Text(str):
    """
    Declares a text node in an HTML document.
    """

    blockLevel = False

    def __new__(cls, obj, escapeMe=True):
        code = obj.encode("utf-8")
        if escapeMe:
            # HTML does not require escaping single quotes but we do this so that
            # the string representation of a document (or document fragment) can
            # be safely enclosed in single quotes.
            code = escape(code).replace("'", "&apos;")
        return str.__new__(cls, code)


class Entity(Text):
    """
    Declares text elements corresponding to the valid (X)HTML entities.
    """

    allowed = [  # http://en.wikipedia.org/wiki/List_of_XML_and_HTML_character_entity_references
        "quot",
        "amp",
        "apos",
        "lt",
        "gt",
        "nbsp",
        "iexcl",
        "cent",
        "pound",
        "curren",
        "yen",
        "brvbar",
        "sect",
        "uml",
        "copy",
        "ordf",
        "laquo",
        "not",
        "shy",
        "reg",
        "macr",
        "deg",
        "plusmn",
        "sup2",
        "sup3",
        "acute",
        "micro",
        "para",
        "middot",
        "cedil",
        "sup1",
        "ordm",
        "raquo",
        "frac14",
        "frac12",
        "frac34",
        "iquest",
        "Agrave",
        "Aacute",
        "Acirc",
        "Atilde",
        "Auml",
        "Aring",
        "AElig",
        "Ccedil",
        "Egrave",
        "Eacute",
        "Ecirc",
        "Euml",
        "Igrave",
        "Iacute",
        "Icirc",
        "Iuml",
        "ETH",
        "Ntilde",
        "Ograve",
        "Oacute",
        "Ocirc",
        "Otilde",
        "Ouml",
        "times",
        "Oslash",
        "Ugrave",
        "Uacute",
        "Ucirc",
        "Uuml",
        "Yacute",
        "THORN",
        "szlig",
        "agrave",
        "aacute",
        "acirc",
        "atilde",
        "auml",
        "aring",
        "aelig",
        "ccedil",
        "egrave",
        "eacute",
        "ecirc",
        "euml",
        "igrave",
        "iacute",
        "icirc",
        "iuml",
        "eth",
        "ntilde",
        "ograve",
        "oacute",
        "ocirc",
        "otilde",
        "ouml",
        "divide",
        "oslash",
        "ugrave",
        "uacute",
        "ucirc",
        "uuml",
        "yacute",
        "thorn",
        "yuml",
        "OElig",
        "oelig",
        "Scaron",
        "scaron",
        "Yuml",
        "fnof",
        "circ",
        "tilde",
        "Alpha",
        "Beta",
        "Gamma",
        "Delta",
        "Epsilon",
        "Zeta",
        "Eta",
        "Theta",
        "Iota",
        "Kappa",
        "Lambda",
        "Mu",
        "Nu",
        "Xi",
        "Omicron",
        "Pi",
        "Rho",
        "Sigma",
        "Tau",
        "Upsilon",
        "Phi",
        "Chi",
        "Psi",
        "Omega",
        "alpha",
        "beta",
        "gamma",
        "delta",
        "epsilon",
        "zeta",
        "eta",
        "theta",
        "iota",
        "kappa",
        "lambda",
        "mu",
        "nu",
        "xi",
        "omicron",
        "pi",
        "rho",
        "sigmaf",
        "sigma",
        "tau",
        "upsilon",
        "phi",
        "chi",
        "psi",
        "omega",
        "thetasym",
        "upsih",
        "piv",
        "ensp",
        "emsp",
        "thinsp",
        "zwnj",
        "zwj",
        "lrm",
        "rlm",
        "ndash",
        "mdash",
        "lsquo",
        "rsquo",
        "sbquo",
        "ldquo",
        "rdquo",
        "bdquo",
        "dagger",
        "Dagger",
        "bull",
        "hellip",
        "permil",
        "prime",
        "Prime",
        "lsaquo",
        "rsaquo",
        "oline",
        "frasl",
        "euro",
        "image",
        "weierp",
        "real",
        "trade",
        "alefsym",
        "larr",
        "uarr",
        "rarr",
        "darr",
        "harr",
        "crarr",
        "lArr",
        "uArr",
        "rArr",
        "dArr",
        "hArr",
        "forall",
        "part",
        "exist",
        "empty",
        "nabla",
        "isin",
        "notin",
        "ni",
        "prod",
        "sum",
        "minus",
        "lowast",
        "radic",
        "prop",
        "infin",
        "ang",
        "and",
        "or",
        "cap",
        "cup",
        "int",
        "there4",
        "sim",
        "cong",
        "asymp",
        "ne",
        "equiv",
        "le",
        "ge",
        "sub",
        "sup",
        "nsub",
        "sube",
        "supe",
        "oplus",
        "otimes",
        "perp",
        "sdot",
        "lceil",
        "rceil",
        "lfloor",
        "rfloor",
        "lang",
        "rang",
        "loz",
        "spades",
        "clubs",
        "hearts",
        "diams",
    ]

    def __new__(cls, tag):
        if tag not in Entity.allowed:
            raise HTMLDocumentError('"%s" is not a recognized HTML entity' % tag)
        return Text.__new__(cls, "&" + tag + ";", escapeMe=False)


class Element(object):
    """
    Declares a generic (X)HTML document element.
    """

    indent = ""

    def __init__(self, *children, **attributes):
        self.tag = self.__class__.__name__.lower()
        if self.tag == "element":
            raise HTMLDocumentError(
                '"element" is a virtual base class for document elements'
            )
        self.root = None
        self.children = []
        for child in children:
            self.append(child)
        self.attributes = {}
        for key in attributes:
            self.setAttribute(key, attributes[key])

    def setAttribute(self, key, value):
        if key == "className":
            key = "class"
        elif not key.islower():
            raise HTMLDocumentError('XHTML attributes must be lowercase: "%s"' % key)
        if key in self.attributes:
            raise HTMLDocumentError(
                "attribute has already been set: %s = %s" % (key, self.attributes[key])
            )
        # We always output attribute values in double quotes so the [&,<,"] characters
        # must be escaped according to http://www.w3.org/TR/REC-xml/#NT-AttValue.
        # We additionally escape single quotes ['] so that an HTML document or
        # fragment can be safely single quoted (but not double quoted).
        safe_value = (
            value.replace('"', "&quot;")
            .replace("'", "&apos;")
            .replace("<", "&gt")
            .replace("&", "&amp;")
        )
        self.attributes[key] = safe_value
        if key == "id" and self.root is not None:
            self.root.registerID(value, self)

    def __getitem__(self, key):
        try:
            return self.attributes[key]
        except KeyError:
            raise HTMLDocumentError(
                '"%s" element has no attribute "%s"' % (self.tag, key)
            )

    def __setitem__(self, key, value):
        self.setAttribute(key, value)

    def setRoot(self, root):
        """
        Assigns or reassigns this element to a document root.

        An element can be added to multiple documents or multiple times to the
        same document (unless it has an 'id' attribute).
        """
        self.root = root
        if "id" in self.attributes and root is not None:
            self.root.registerID(self.attributes["id"], self)
        for child in self.children:
            if isinstance(child, Element):
                child.setRoot(root)

    def append(self, child):
        if not isinstance(child, Element) and not isinstance(child, Text):
            child = Text(child)
        self.children.append(child)
        if isinstance(child, Element) and self.root is not None:
            child.setRoot(self.root)

    def extend(self, children):
        for child in children:
            self.append(child)

    def __str__(self):
        """
        Return a string representation of this element.

        Note that the returned string will contain double quotes around all attribute
        values but will not contain any single quotes (except possibly via an
        un-escaped text node), so can be safely enclosed in single quotes.
        """
        s = ""
        if self.blockLevel:
            s += Element.indent
        s += "<%s" % self.tag
        for attrib in self.attributes:
            s += ' %s="%s"' % (attrib, self.attributes[attrib])
        if len(self.children) == 0:
            s += " />"
        else:
            s += ">"
            blocked = True in [child.blockLevel for child in self.children]
            if self.blockLevel and blocked:
                s += "\n"
                Element.indent += "  "
            for child in self.children:
                s += child.__str__()
            if self.blockLevel and blocked:
                Element.indent = Element.indent[:-2]
                s += Element.indent
            s += "</%s>" % self.tag
        if self.blockLevel:
            s += "\n"
        return s


# Define the allowed HTML4 block-level elements according to
# http://htmlhelp.com/reference/html40/block.html
# Python symbols are capitalized to avoid conflicts with builtin names (like 'object')
for tag in [
    # defined as block level in HTML4
    "address",
    "blockquote",
    "div",
    "dl",
    "fieldset",
    "form",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "hr",
    "noscript",
    "ol",
    "p",
    "pre",
    "table",
    "ul",
    # are considered block level since they can contain block-level children
    "dd",
    "dt",
    "li",
    "tbody",
    "td",
    "tfoot",
    "th",
    "thead",
    "tr",
    # are block level when they contain block-level children
    "button",
    "del",
    "ins",
    "map",
    "object",
    "script",
    # not listed as either block-level or inline at htmlhelp.com
    "body",
    # not HTML4 standard
    "embed",
]:
    globals()[tag.capitalize()] = type(
        tag.capitalize(), (Element,), {"blockLevel": True}
    )

# Define the allowed HTML4 inline elements according to
# http://htmlhelp.com/reference/html40/inline.html
# Python symbols are capitalized to avoid conflicts with builtin names (like 'object')
for tag in [
    # defined as inline in HTML4
    "a",
    "abbr",
    "acronym",
    "b",
    "bdo",
    "big",
    "br",
    "cite",
    "code",
    "dfn",
    "em",
    "i",
    "img",
    "input",
    "kbd",
    "label",
    "q",
    "samp",
    "select",
    "small",
    "span",
    "strong",
    "sub",
    "sup",
    "textarea",
    "tt",
    "var",
    # not listed as either block-level or inline at htmlhelp.com
    "option",
    "optgroup",
]:
    globals()[tag.capitalize()] = type(
        tag.capitalize(), (Element,), {"blockLevel": False}
    )


class HTMLDocument:
    """
    Declares an (X)HTML document.
    """

    def __init__(self, head, body):
        self.head = head
        self.body = body
        self.registry = {}
        self.body.setRoot(self)

    def registerID(self, id, element):
        if id in self.registry:
            raise HTMLDocumentError(
                'id already registered for this document: "%s"' % id
            )
        self.registry[id] = element

    def __getitem__(self, key):
        try:
            return self.registry[key]
        except KeyError:
            raise HTMLDocumentError(
                'Document has no element registered with id "%s"' % key
            )

    def __str__(self):
        return (
            """<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
  "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">\n"""
            + self.head.__str__()
            + self.body.__str__()
            + "\n</html>"
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-05-20
# @Filename: base.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import abc
import asyncio
import inspect
import logging
import pathlib
import time
from datetime import datetime

from typing import Any, Dict, Optional, TypeVar, Union, cast

from sdsstools import get_logger, read_yaml_file
from sdsstools.logger import SDSSLogger

from clu.command import BaseCommand

from .model import Model
from .tools import REPLY


__all__ = ["BaseClient", "BaseActor", "Reply"]

SchemaType = Union[Dict[str, Any], pathlib.Path, str]
T = TypeVar("T", bound="BaseClient")


class BaseClient(metaclass=abc.ABCMeta):
    """A base client that can be used for listening or for an actor.

    This class defines a new client. Clients differ from actors in that
    they do not receive commands or issue replies, but do send commands to
    other actors and listen to the keyword-value flow. All actors are also
    clients and any actor should subclass from `.BaseClient`.

    Normally a new instance of a client or actor is created by passing a
    configuration file path to `.from_config` which defines how the
    client must be started.

    Parameters
    ----------
    name
        The name of the client.
    version
        The version of the client.
    loop
        The event loop. If `None`, the current event loop will be used.
    log_dir
        The directory where to store the file logs.
    log
        A `~logging.Logger` instance to be used for logging instead of creating
        a new one.
    verbose
        Whether to log to stdout. Can be an integer logging level.
    validate
        Whether to actor should validate its own messages against its model (if it
        has one). This is a global parameter that can be overridden when calling
        `~.BaseClient.write`.
    """

    name: str

    def __init__(
        self,
        name: str,
        version: Optional[str] = None,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        log_dir: Optional[Union[pathlib.Path, str]] = None,
        log: Optional[SDSSLogger] = None,
        verbose: Union[bool, int] = False,
        validate: bool = True,
    ):

        self.loop = loop or asyncio.get_event_loop()

        self.name = name
        assert self.name, "name cannot be empty."

        self.log: SDSSLogger
        self.setup_logger(log, log_dir, verbose=verbose)

        self.version = version or "?"

        # Internally store the original configuration used to start the client.
        self.config: Dict[str, Any] = {}

        self.validate = validate

    def __repr__(self):
        return f"<{str(self)} (name={self.name!r})>"

    def __str__(self):
        return self.__class__.__name__

    @abc.abstractmethod
    async def start(self: T) -> T:
        """Runs the client."""

        pass

    async def stop(self):
        """Shuts down all the remaining tasks."""

        self.log.info("cancelling all pending tasks and shutting down.")

        tasks = [
            task
            for task in asyncio.all_tasks(loop=self.loop)
            if task is not asyncio.current_task(loop=self.loop)
        ]
        list(map(lambda task: task.cancel(), tasks))

        await asyncio.gather(*tasks, return_exceptions=True)

        self.loop.stop()

    @staticmethod
    def _parse_config(
        input: Union[Dict[str, Any], pathlib.Path, str]
    ) -> Dict[str, Any]:

        if not isinstance(input, dict):
            input = pathlib.Path(input)
            assert input.exists(), "configuration path does not exist."
            config = read_yaml_file(str(input))
        else:
            config = input

        return cast("Dict[str, Any]", config)

    @classmethod
    def from_config(
        cls,
        config: Union[Dict[str, Any], pathlib.Path, str],
        *args,
        **kwargs,
    ):
        """Parses a configuration file.

        Parameters
        ----------
        config
            A configuration dictionary or the path to a YAML configuration
            file. If the file contains a section called ``'actor'`` or
            ``'client'``, that section will be used instead of the whole
            file.
        """

        orig_config_dict = cls._parse_config(config)

        config_dict = orig_config_dict.copy()

        if "actor" in config_dict:
            config_dict = config_dict["actor"]
        elif "client" in config_dict:
            config_dict = config_dict["client"]

        config_dict.update(kwargs)

        # Decide what to do with the rest of the keyword arguments:
        args_inspect = inspect.getfullargspec(cls)

        if args_inspect.varkw is not None:
            # If there is a catch-all kw variable, send everything and let the
            # subclass handle it.
            config_dict.update(kwargs)
        else:
            # Check the kw arguments in the subclass and pass only
            # values from config_dict that match them.
            class_kwargs = args_inspect.args
            class_kwargs.remove("self")

            # Remove keys that are not in the signature.
            config_dict = {
                key: value for key, value in config_dict.items() if key in class_kwargs
            }

        # We also pass *args in case the actor has been subclassed
        # and the subclass' __init__ accepts different arguments.
        new_client = cls(*args, **config_dict)

        # Store original config. This may not be complete since from_config
        # may have been super'd from somewhere else.
        new_client.config = orig_config_dict

        return new_client

    def setup_logger(
        self,
        log: Any,
        log_dir: Optional[Union[pathlib.Path, str]],
        verbose: Union[bool, int] = False,
    ):
        """Starts the file logger."""

        if not log:
            log = get_logger("clu:" + self.name)
        else:
            assert isinstance(log, SDSSLogger), "Logger must be sdsstools.SDSSLogger"
            self.log = log
            return log

        log.setLevel(REPLY)

        if log is not False and log_dir:

            log_dir = pathlib.Path(log_dir).expanduser()

            log.start_file_logger(
                str(log_dir / f"{self.name}.log"),
                rotating=True,
                rollover=True,
            )

            if log.fh:  # In case starting the file logger fails.
                log.fh.formatter.converter = time.gmtime
                log.fh.setLevel(REPLY)

        log.sh.setLevel(logging.WARNING)
        if verbose is True:
            log.sh.setLevel(logging.DEBUG)
        elif verbose is not False and isinstance(verbose, int):
            log.sh.setLevel(verbose)

        self.log = log
        self.log.debug(f"{self.name}: logging system initiated.")

        # Set the loop exception handler to be handled by the logger.
        self.loop.set_exception_handler(self.log.asyncio_exception_handler)

        return log

    def send_command(self, actor: str, *args, **kwargs):  # pragma: no cover
        """Sends a command to an actor and returns a `.Command` instance."""

        raise NotImplementedError(
            "Sending commands is not implemented for this client."
        )


class BaseActor(BaseClient):
    """An actor based on `asyncio`.

    This class expands `.BaseClient` with a parsing system for new commands
    and placeholders for methods for handling new commands and writing replies,
    which should be overridden by the specific actors.
    """

    model: Union[Model, None] = None

    def __init__(self, *args, schema: SchemaType = None, **kwargs):

        super().__init__(*args, **kwargs)

        self.load_schema(schema)

    def load_schema(
        self,
        schema: Union[SchemaType, None],
        is_file=True,
    ) -> Union[Model, None]:
        """Loads and validates the actor schema."""

        if schema is None:
            schema = {
                "$schema": "http://json-schema.org/draft-07/schema#",
                "type": "object",
                "properties": {},
                "additionalProperties": True,
            }
            is_file = False

        self.model = Model(self.name, schema, is_file=is_file, log=self.log)

        return self.model

    @abc.abstractmethod
    def new_command(self):
        """Handles a new command.

        Must be overridden by the subclass and call `.parse_command`
        with a `.Command` object.
        """
        pass

    @abc.abstractmethod
    def parse_command(self, command: BaseCommand):
        """Parses and executes a `.Command`. Must be overridden."""
        pass

    def send_command(self):
        """Sends a command to another actor."""

        raise NotImplementedError("Sending commands is not implemented for this actor.")

    @abc.abstractmethod
    def _write_internal(self, reply: Reply):
        """Internally handle the reply and output it to the users.

        Must handle converting the general `.Reply` to the specific format of the
        actor transport. Must also handle logging the reply.
        """
        pass

    def write(
        self,
        message_code: str = "i",
        message: Optional[Dict[str, Any] | str] = None,
        command: Optional[BaseCommand] = None,
        broadcast: bool = False,
        validate: bool | None = None,
        call_internal: bool = True,
        **kwargs,
    ) -> Reply:
        """Writes a message to user(s).

        The reply to the users will be formatted by the actor class into message
        specific to the communication channel. Additional keywords passed to this
        method will be used to complete the message (as long as they don't overlap
        with named parameters). For example ::

            actor.write('i', message={'text': 'Hi!', 'value1': 1})

        and ::

            actor.write('i', text='Hi!', value1=1)

        are equivalent and ::

            actor.write('i', message={'text': 'Hi!'}, value1=1)

        is equivalent to ::

            actor.write('i', message={'text': 'Hi!', 'value1': 1})

        This method generates a `.Reply` object that is passed to the
        ``_write_internal`` method in the class, which processes it and outputs the
        message to the users using the appropriate transport. If ``command`` is passed,
        the reply is added to ``command.replies``.

        Parameters
        ----------
        message_code
            The message code (e.g., ``'i'`` or ``':'``).
        message
            The keywords to be output. Must be a dictionary of pairs
            ``{keyword: value}``.
        command
            The command to which we are replying. If not set, it is assumed
            that this is a broadcast.
        broadcast
            Whether to broadcast the message to all the users or only to the
            commander.
        validate
            Validate the reply against the actor schema. This is ignored if the actor
            was not started with knowledge of its own schema. If `None`, defaults to
            the actor global behaviour.
        call_internal
            Whether to call the actor internal write method. Should be `True` but
            it's sometimes useful to call `.write` with ``call_internal=False`` when
            one is overriding the method and wants to control when to call the internal
            method.
        kwargs
            Keyword arguments that will used to update the message.
        """

        if isinstance(message, str):
            message = {"text": message}
        else:
            message = message or {}
            if not isinstance(message, dict):
                raise TypeError("message must be a dictionary")

        message.update(kwargs)

        reply = Reply(message_code, message, command=command, broadcast=broadcast)

        do_validate = validate if validate is not None else self.validate
        if do_validate and self.model is not None:
            reply.use_validation = True
            result, err = self.model.update_model(message)
            if result is False:
                message = {"error": f"Failed validating the reply: {err}"}
                reply.message_code = "e"
                reply.message = message
                reply.validated = False
            else:
                reply.validated = True

        if command:
            command.replies.append(reply)

        if call_internal:
            if asyncio.iscoroutinefunction(self._write_internal):
                asyncio.create_task(self._write_internal(reply))  # type: ignore
            else:
                self._write_internal(reply)

        return reply


class Reply:
    """A reply from a command or actor to be sent to the users."""

    def __init__(
        self,
        message_code: str,
        message: Dict[str, Any],
        command: Optional[BaseCommand] = None,
        broadcast: bool = False,
        use_validation: bool = False,
        validated: bool = False,
    ):
        self.date = datetime.utcnow()
        self.message_code = message_code
        self.message = message
        self.command = command
        self.broadcast = broadcast
        self.use_validation = use_validation
        self.validated = validated



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2018-09-07
# @Filename: tools.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import contextlib
import enum
import functools
import inspect
import json
import logging
import re

from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
)


__all__ = [
    "CommandStatus",
    "StatusMixIn",
    "format_value",
    "CallbackMixIn",
    "CaseInsensitiveDict",
    "cli_coro",
    "as_complete_failer",
    "log_reply",
    "ActorHandler",
]

REPLY = 5  # REPLY logging level
WARNING_REGEX = r"^.*?\s*?(\w*?Warning): (.*)"


class Maskbit(enum.Flag):
    """A maskbit enumeration. Intended for subclassing."""

    @property
    def active_bits(self) -> List[Maskbit]:
        """Returns a list of non-combination flags that match the value."""

        return [
            bit
            for bit in self.__class__  # type: ignore
            if ((bit.value & self.value) and bin(bit.value).count("1") == 1)
        ]


COMMAND_STATUS_TO_CODE: Dict[str, str] = {
    "DONE": ":",
    "CANCELLED": "f",
    "FAILED": "f",
    "TIMEDOUT": "f",
    "READY": "i",
    "RUNNING": ">",
    "CANCELLING": "w",
    "FAILING": "w",
    "DEBUG": "d",
}


class CommandStatus(Maskbit):

    DONE = enum.auto()
    CANCELLED = enum.auto()
    FAILED = enum.auto()
    TIMEDOUT = enum.auto()
    READY = enum.auto()
    RUNNING = enum.auto()
    CANCELLING = enum.auto()
    FAILING = enum.auto()
    DEBUG = enum.auto()

    ACTIVE_STATES = RUNNING | CANCELLING | FAILING
    FAILED_STATES = CANCELLED | FAILED | TIMEDOUT
    FAILING_STATES = CANCELLING | FAILING
    DONE_STATES = DONE | FAILED_STATES
    ALL_STATES = READY | ACTIVE_STATES | DONE_STATES

    def __init__(self, *args):

        if self.name.upper() in COMMAND_STATUS_TO_CODE:
            self.code = COMMAND_STATUS_TO_CODE[self.name.upper()]
        else:
            self.code = None

    @property
    def is_combination(self) -> bool:
        """Returns True if a flag is a combination."""

        if bin(self.value).count("1") > 1:
            return True
        return False

    @property
    def did_fail(self) -> bool:
        """Command failed or was cancelled."""

        return self in self.FAILED_STATES

    @property
    def did_succeed(self) -> bool:
        """Command finished with DONE status."""

        return self == self.DONE

    @property
    def is_active(self) -> bool:
        """Command is running, cancelling or failing."""

        return self in self.ACTIVE_STATES

    @property
    def is_done(self) -> bool:
        """Command is done (whether successfully or not)."""

        return self in self.DONE_STATES

    @property
    def is_failing(self) -> bool:
        """Command is being cancelled or is failing."""

        return self in self.FAILING_STATES

    @staticmethod
    def code_to_status(code, default: Optional[CommandStatus] = None) -> CommandStatus:
        """Returns the status associated with a code.

        If the code doesn't have an associated status, returns ``default``.
        ``default`` defaults to `.CommandStatus.RUNNING`.

        """

        statuses = {
            ":": CommandStatus.DONE,
            "f": CommandStatus.FAILED,
            "e": CommandStatus.FAILED,
            "!": CommandStatus.FAILED,
            ">": CommandStatus.RUNNING,
        }

        return statuses.get(code, default or CommandStatus.RUNNING)


MaskbitType = TypeVar("MaskbitType", bound=Maskbit)


class StatusMixIn(Generic[MaskbitType]):
    """A mixin that provides status tracking with callbacks.

    Provides a status property that executes a list of callbacks when
    the status changes.

    Parameters
    ----------
    maskbit_flags
        A class containing the available statuses as a series of maskbit
        flags. Usually as subclass of `enum.Flag`.
    initial_status
        The initial status.
    callback_func
        The function to call if the status changes.
    call_now
        Whether the callback function should be called when initialising.

    Attributes
    ----------
    callbacks
        A list of the callback functions to call.
    """

    def __init__(
        self,
        maskbit_flags: Type[MaskbitType],
        initial_status: Optional[MaskbitType] = None,
        callback_func: Optional[Callable[[MaskbitType], Any]] = None,
        call_now: bool = False,
    ):

        self.flags = maskbit_flags
        self.callbacks: List[Callable[[MaskbitType], Any]] = []
        self._status: MaskbitType | None = initial_status
        self.watcher: Optional[asyncio.Event] = None

        if callback_func is not None:
            if isinstance(callback_func, (list, tuple)):
                self.callbacks = callback_func
            else:
                self.callbacks.append(callback_func)

        if call_now is True:
            self.do_callbacks()

    def do_callbacks(self):
        """Calls functions in ``callbacks``."""

        assert hasattr(self, "callbacks"), "missing callbacks attribute."

        loop = asyncio.get_event_loop()

        for func in self.callbacks:
            loop.call_soon(func)

    @property
    def status(self):
        """Returns the status."""

        return self._status

    @status.setter
    def status(self, value):
        """Sets the status."""

        if value != self._status:
            self._status = value
            self.do_callbacks()
            if self.watcher is not None:
                self.watcher.set()

    async def wait_for_status(self, value):
        """Awaits until the status matches ``value``."""

        if self.status == value:
            return

        self.watcher = asyncio.Event()

        while self.status != value:
            await self.watcher.wait()
            if self.watcher is not None:
                self.watcher.clear()

        self.watcher = None


class CallbackMixIn(object):
    """A mixin for executing callbacks.

    Parameters
    ----------
    callbacks
        A list of functions or coroutines to be called.
    """

    def __init__(
        self,
        callbacks: List[Callable[[Any], Any]] = [],
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ):

        self._callbacks = []
        for cb in callbacks:
            self.register_callback(cb)

        self._running = []  # Running callbacks

        self.loop = loop or asyncio.get_event_loop()

    async def stop_callbacks(self):
        """Cancels any running callback task."""

        for cb in self._running:
            if not cb.done():
                cb.cancel()

        with contextlib.suppress(asyncio.CancelledError):
            for cb in self._running:
                await cb

        self._running = []

    def register_callback(self, callback_func: Callable[..., Any]):
        """Adds a callback function or coroutine function."""

        assert callable(callback_func), "callback_func must be a callable."
        self._callbacks.append(callback_func)

    def remove_callback(self, callback_func: Callable[..., Any]):
        """Removes a callback function."""

        assert (
            callback_func in self._callbacks
        ), "callback_func is not in the list of callbacks."
        self._callbacks.remove(callback_func)

    def notify(self, *args):
        """Calls the callback functions with some arguments.

        Coroutine callbacks are scheduled as a task. Synchronous callbacks
        are scheduled with ``call_soon``.

        """

        if self._callbacks is None:
            return

        for cb in self._callbacks:
            n_args = len(inspect.getfullargspec(cb).args)
            if asyncio.iscoroutinefunction(cb):
                task = asyncio.create_task(cb(*args[:n_args]))
                self._running.append(task)
                # Auto-dispose of the task once it completes
                task.add_done_callback(self._running.remove)
            else:
                # Check that the loop is running. There is a problem in which
                # self.loop may be set before there is a running loop so we
                # replace it with a properly running loop.
                if not self.loop or self.loop.is_running() is False:
                    self.loop = asyncio.get_event_loop()
                task = self.loop.call_soon(cb, *args[:n_args])


def dict_depth(d: dict) -> int:
    """Gets the depth of a dictionary."""
    if isinstance(d, dict):
        return 1 + (max(map(dict_depth, d.values())) if d else 0)
    return 0


def format_value(value: Any) -> str:
    """Formats messages in a way that is compatible with the parser.

    Parameters
    ----------
    value
        The data to be formatted.

    Returns
    -------
    formatted_text
        A string with the escaped text.
    """

    if isinstance(value, str):
        if " " in value and not (value.startswith("'") or value.startswith('"')):
            value = escape(value)
    elif isinstance(value, bool):
        value = "T" if value else "F"
    elif isinstance(value, (tuple, list)):
        value = ",".join([format_value(item) for item in value])
    elif isinstance(value, dict):
        if dict_depth(value) > 1:
            raise ValueError("Cannot format a dictionary with depth > 1.")
        value = format_value(list(value.values()))
    else:
        value = str(value)

    return value


def escape(value: Any):
    """Escapes a text using `json.dumps`."""

    return json.dumps(value)


T = TypeVar("T")


class CaseInsensitiveDict(Dict[str, T]):
    """A dictionary that performs case-insensitive operations."""

    def __init__(self, values: Any):

        self._lc = []

        dict.__init__(self, values)

        self._lc = [key.lower() for key in values]
        assert len(set(self._lc)) == len(
            self._lc
        ), "the are duplicated items in the dict."

    def __get_key__(self, key):
        """Returns the correct value of the key, regardless of its case."""

        try:
            idx = self._lc.index(key.lower())
        except ValueError:
            return key

        return list(self)[idx]

    def __getitem__(self, key):
        return dict.__getitem__(self, self.__get_key__(key))

    def __setitem__(self, key, value):

        if key.lower() not in self._lc:
            self._lc.append(key.lower())
            dict.__setitem__(self, key, value)
        else:
            dict.__setitem__(self, self.__get_key__(key), value)

    def __contains__(self, key):
        return dict.__contains__(self, self.__get_key__(key))

    def __eq__(self, key):
        return dict.__eq__(self, self.__get_key__(key))


def cli_coro(f):
    """Decorator function that allows defining coroutines with click."""

    f = asyncio.coroutine(f)

    def wrapper(*args, **kwargs):
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(f(*args, **kwargs))

    return functools.update_wrapper(wrapper, f)


async def as_complete_failer(
    aws: List[Awaitable],
    on_fail_callback: Optional[Callable] = None,
    **kwargs,
) -> Tuple[bool, str | None]:
    """Similar to `~asyncio.as_complete` but cancels all the tasks
    if any of them returns `False`.

    Parameters
    ----------
    aws
        A list of awaitable objects. If not a list, it will be wrapped in one.
    on_fail_callback
        A function or coroutine to call if any of the tasks failed.
    kwargs
        A dictionary of keywords to be passed to `~asyncio.as_complete`.

    Returns
    -------
    result_tuple
        A tuple in which the first element is `True` if all the tasks
        completed, `False` if any of them failed and the rest were cancelled.
        If `False`, the second element is `None` if no exceptions were caught
        during the execution of the tasks, otherwise it contains the error
        message. If `True`, the second element is always `None`.
    """

    if not isinstance(aws, (list, tuple)):
        aws = [aws]

    loop = kwargs.get("loop", asyncio.get_event_loop())

    tasks = [loop.create_task(aw) for aw in aws]

    failed = False
    error_message = None
    for next_completed in asyncio.as_completed(tasks, **kwargs):
        try:
            result = await next_completed
        except Exception as ee:
            error_message = str(ee)
            result = False

        if not result:
            failed = True
            break

    if failed:

        # Cancel tasks
        [task.cancel() for task in tasks]

        with contextlib.suppress(BaseException):
            await asyncio.gather(*[task for task in tasks])

        if on_fail_callback:
            if asyncio.iscoroutinefunction(on_fail_callback):
                await on_fail_callback()
            else:
                on_fail_callback()

        return (False, error_message)

    return (True, None)


def log_reply(
    log: logging.Logger,
    message_code: str,
    message: str,
    use_message_code: bool = False,
):
    """Logs an actor message with the correct code."""

    code_dict = {
        "f": logging.ERROR,
        "e": logging.ERROR,
        "w": logging.WARNING,
        "i": logging.INFO,
        ":": logging.INFO,
        "d": logging.DEBUG,
    }

    if use_message_code:
        log.log(code_dict[message_code], message)
    else:
        # Sets the REPLY log level
        log_level_no = REPLY
        if log_level_no in logging._levelToName:
            log_level = log_level_no
        else:
            log_level = logging.DEBUG

        log.log(log_level, message)


_ActorClass = TypeVar("_ActorClass")


class ActorHandler(logging.Handler):
    """A handler that outputs log messages as actor keywords.

    Parameters
    ----------
    actor
        The actor instance.
    level
        The level above which records will be output in the actor.
    keyword
        The keyword around which the messages will be output.
    code_mapping
        A mapping of logging levels to actor codes. The values provided
        override the default mapping. For example, to make input log messages
        with info level be output as debug,
        ``code_mapping={logging.INFO: 'd'}``.
    filter_warnings
        A list of warning classes that will be issued to the actor. Subclasses
        of the filter warning are accepted, any other warnings will be ignored.
    """

    def __init__(
        self,
        actor,
        level: int = logging.ERROR,
        keyword: str = "text",
        code_mapping: Optional[Dict[int, str]] = None,
        filter_warnings: Optional[List[Type[Warning]]] = None,
    ):

        self.actor = actor
        self.keyword = keyword

        self.code_mapping = {
            logging.DEBUG: "d",
            logging.INFO: "i",
            logging.WARNING: "w",
            logging.ERROR: "f",
        }

        if code_mapping:
            self.code_mapping.update(code_mapping)

        self.filter_warnings = filter_warnings

        super().__init__(level=level)

    def emit(self, record: logging.LogRecord):
        """Emits the record."""

        message = record.getMessage()
        message_lines = message.splitlines()

        if record.exc_info:
            message_lines.append(f"{record.exc_info[0].__name__}: {record.exc_info[1]}")

        if record.levelno <= logging.DEBUG:
            code = self.code_mapping[logging.DEBUG]
        elif record.levelno <= logging.INFO:
            code = self.code_mapping[logging.INFO]
        elif record.levelno <= logging.WARNING:
            code = self.code_mapping[logging.WARNING]
            warning_category_groups = re.match(WARNING_REGEX, message)
            if warning_category_groups is not None:
                message_lines = self._filter_warning(warning_category_groups)
        elif record.levelno >= logging.ERROR:
            code = self.code_mapping[logging.ERROR]
        else:
            code = "w"

        for line in message_lines:
            result = self.actor.write(code, message={self.keyword: line})

            if asyncio.iscoroutine(result):
                asyncio.create_task(result)

    def _filter_warning(self, warning_category_groups):

        warning_category, warning_text = warning_category_groups.groups()
        message_lines = [f"{warning_text} ({warning_category})"]

        try:
            if self.filter_warnings:
                for warning_filter in self.filter_warnings:
                    if warning_category == warning_filter.__name__:
                        return message_lines
            return []
        except NameError:
            return message_lines



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-07-30
# @Filename: client.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import json
import logging
import pathlib
import uuid

from typing import Any, Dict, List, Optional, Union

import aio_pika as apika

from sdsstools.logger import SDSSLogger

from .base import BaseClient
from .command import Command
from .model import ModelSet
from .protocol import TopicListener
from .tools import CommandStatus


__all__ = ["AMQPClient", "AMQPReply"]


PathLike = Union[str, pathlib.Path]


class AMQPReply(object):
    """Wrapper for an `~aio_pika.IncomingMessage` that expands and decodes it.

    Parameters
    ----------
    message
        The message that contains the reply.
    log
        A message logger.

    Attributes
    ----------
    is_valid
        Whether the message is valid and correctly parsed.
    body
        The body of the message, as a JSON dictionary.
    info
        The info dictionary.
    headers
        The headers of the message, decoded if they are bytes.
    message_code
        The message code.
    sender
        The name of the actor that sends the reply.
    command_id
        The command ID.
    """

    def __init__(
        self,
        message: apika.IncomingMessage,
        log: Optional[logging.Logger] = None,
    ):

        self.message = message
        self.log = log

        self.is_valid = True

        self.body = None

        # Acknowledges receipt of message
        message.ack()

        self.info: Dict[Any, Any] = message.info()

        self.headers = self.info["headers"]
        for key in self.headers:
            if isinstance(self.headers[key], bytes):
                self.headers[key] = self.headers[key].decode()

        self.message_code = self.headers.get("message_code", None)

        if self.message_code is None:
            self.is_valid = False
            if self.log:
                self.log.warning(f"received message without message_code: {message}")
            return

        self.sender = self.headers.get("sender", None)
        if self.sender is None and self.log:
            self.log.warning(f"received message without sender: {message}")

        self.command_id = message.correlation_id

        command_id_header = self.headers.get("command_id", None)
        if command_id_header and command_id_header != self.command_id:
            if self.log:
                self.log.error(
                    f"mismatch between message "
                    f"correlation_id={self.command_id} "
                    f"and header command_id={command_id_header} "
                    f"in message {message}"
                )
            self.is_valid = False
            return

        self.body = json.loads(self.message.body.decode())


class AMQPClient(BaseClient):
    """Defines a new client based on the AMQP standard.

    To start a new client first instantiate the class and then run `.start` as
    a coroutine. Note that `.start` does not block so you will need to use
    asyncio's ``run_forever`` or a similar system ::

        >>> loop = asyncio.get_event_loop()
        >>> client = await AMQPClient('my_client', host='localhost').start()
        >>> loop.run_forever()

    Parameters
    ----------
    name
        The name of the client.
    url
        RFC3986 formatted broker address. When used, the other connection
        keyword arguments are ignored.
    user
        The user to connect to the AMQP broker. Defaults to ``guest``.
    password
        The password for the user. Defaults to ``guest``.
    host
        The host where the AMQP message broker runs. Defaults to ``localhost``.
    virtualhost
         Virtualhost parameter. ``'/'`` by default.
    port
        The port on which the AMQP broker is running. Defaults to 5672.
    ssl
        Whether to use TLS/SSL connection.
    version
        The version of the client.
    loop
        The event loop. If `None`, the current event loop will be used.
    log_dir
        The directory where to store the logs. Defaults to
        ``$HOME/logs/<name>`` where ``<name>`` is the name of the actor.
    log
        A `~logging.Logger` instance to be used for logging instead of
        creating a new one.
    parser
        A click command parser that is a subclass of `~clu.parser.CluGroup`.
        If `None`, the active parser will be used.
    models
        A list of actor models whose schemas will be monitored.
    """

    __EXCHANGE_NAME__ = "sdss_exchange"

    connection = None

    def __init__(
        self,
        name: str,
        url: Optional[str] = None,
        user: str = "guest",
        password: str = "guest",
        host: str = "localhost",
        port: int = 5672,
        virtualhost: str = "/",
        ssl: bool = False,
        version: Optional[str] = None,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        log_dir: Optional[PathLike] = None,
        log: Optional[SDSSLogger] = None,
        models: List[str] = [],
    ):

        super().__init__(name, version=version, loop=loop, log_dir=log_dir, log=log)

        self.replies_queue = None

        # Creates the connection to the AMQP broker
        self.connection = TopicListener(
            url=url,
            user=user,
            password=password,
            host=host,
            port=port,
            ssl=ssl,
            virtualhost=virtualhost,
        )

        #: dict: External commands currently running.
        self.running_commands = {}

        self.models = ModelSet(self, actors=models, raise_exception=False, log=self.log)

    def __repr__(self):

        if not self.connection or self.connection.connection is None:
            url = "disconnected"
        else:
            url = str(self.connection.connection.url)

        return f"<{str(self)} (name={self.name!r}, {url}>"

    async def start(self, exchange_name: str = __EXCHANGE_NAME__):
        """Starts the connection to the AMQP broker."""

        # Starts the connection and creates the exchange
        await self.connection.connect(exchange_name)

        # Binds the replies queue.
        self.replies_queue = await self.connection.add_queue(
            f"{self.name}_replies",
            callback=self.handle_reply,
            bindings=["reply.#"],
        )

        self.log.info(
            f"replies queue {self.replies_queue.name!r} "
            f"bound to {self.connection.connection.url!s}"
        )

        # Initialises the models.
        await self.models.load_schemas()

        return self

    async def stop(self):
        """Cancels queues and closes the connection."""

        await self.connection.stop()

    async def run_forever(self):
        """Runs the event loop forever."""

        while not self.connection.connection.is_closed:
            await asyncio.sleep(1)

    async def handle_reply(self, message: apika.IncomingMessage) -> AMQPReply:
        """Handles a reply received from the exchange.

        Creates a new instance of `.AMQPReply` from the ``message``. If the
        reply is valid it updates any running command.

        Parameters
        ----------
        message
            The message received.

        Returns
        -------
        reply
            The `.AMQPReply` object created from the message.
        """

        reply = AMQPReply(message, log=self.log)

        if not reply.is_valid:
            self.log.error("Invalid message received.")
            return reply

        # Ignores message from self, because actors are also clients and they
        # receive their own messages. The exception is when the commander is also the
        # actor (an actor sent a command to itself).
        commander_id = reply.headers["commander_id"]
        if reply.sender and self.name == reply.sender:
            if commander_id and commander_id != self.name:
                return reply

        # Update the models
        if self.models and reply.sender in self.models:
            self.models[reply.sender].update_model(reply.body)

        # If the command is running we check if the message code indicates
        # the command is done and, if so, sets the result in the Future.
        # Also, add the reply to the command list of replies.
        if reply.command_id in self.running_commands:
            self.running_commands[reply.command_id].replies.append(reply)
            status = CommandStatus.code_to_status(reply.message_code)
            if status.is_done:
                command = self.running_commands.pop(reply.command_id)
                command.set_status(status)
                if not command.done():
                    command.set_result(command)

        return reply

    async def send_command(
        self,
        consumer: str,
        command_string: str,
        command_id: Optional[str] = None,
    ):
        """Commands another actor over its RCP queue.

        Parameters
        ----------
        consumer
            The actor we are commanding.
        command_string
            The command string that will be parsed by the remote actor.
        command_id
            The command ID associated with this command. If empty, an unique
            identifier will be attached.
        """

        command_id = command_id or str(uuid.uuid4())

        # Creates and registers a command.
        command = Command(
            command_string=command_string,
            command_id=command_id,
            commander_id=self.name,
            consumer_id=consumer,
            actor=None,
            loop=self.loop,
        )

        self.running_commands[command_id] = command

        headers = {"command_id": command_id, "commander_id": self.name}

        # The routing key has the topic command and the name of
        # the commanded actor.
        routing_key = f"command.{consumer}"

        message_body = {"command_string": command_string}

        try:
            await self.connection.exchange.publish(
                apika.Message(
                    json.dumps(message_body).encode(),
                    content_type="text/json",
                    headers=headers,
                    correlation_id=command_id,
                    reply_to=self.replies_queue.name,
                ),
                routing_key=routing_key,
            )
        except (apika.exceptions.DeliveryError, apika.exceptions.PublishError):
            # The consumer (actor) did not reply. This usually means that the actor
            # is not connected. We fake a reply from that actor saying so. That will
            # be received by handle_reply which will fail the current command.
            error_msg = dict(error=f"Failed routing message to consumer {consumer!r}.")
            headers.update({"message_code": "f", "sender": consumer})
            await self.connection.exchange.publish(
                apika.Message(
                    json.dumps(error_msg).encode(),
                    content_type="text/json",
                    headers=headers,
                    correlation_id=command_id,
                ),
                routing_key=f"reply.{command.commander_id}",
            )

        return command



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-05-17
# @Filename: model.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import json
import pathlib
import warnings
from os import PathLike

from typing import Any, Callable, Dict, List, Optional, TypeVar, Union, cast

import jsonschema
import jsonschema.exceptions
import jsonschema.validators

import clu.base

from .exceptions import CluError, CluWarning
from .tools import CallbackMixIn, CaseInsensitiveDict


__all__ = ["Property", "BaseModel", "Model", "ModelSet"]


SchemaType = Union[Dict[str, Any], PathLike, str]

DEFAULT_SCHEMA = {
    "text": {"type": "string"},
    "schema": {"type": "string"},
    "version": {"type": "string"},
    "help": {
        "oneOf": [
            {"type": "array", "items": {"type": "string"}},
            {"type": "string"},
        ]
    },
    "error": {
        "oneOf": [
            {"type": "array", "items": {"type": "string"}},
            {"type": "string"},
        ]
    },
    "yourUserID": {"type": "integer"},
    "UserInfo": {
        "type": "array",
        "items": [{"type": "integer"}, {"type": "string"}],
    },
    "num_users": {"type": "integer"},
}


class Property(CallbackMixIn):
    """A model property with callbacks.

    Parameters
    ----------
    name
        The name of the property.
    value
        The value of the property.
    model
        The parent model.
    callback
        The function or coroutine that will be called if the value of the key
        if updated. The callback is called with the instance of `Property`
        as the only argument. Note that the callback will be scheduled even
        if the value does not change.
    """

    def __init__(
        self,
        name: str,
        value: Optional[Any] = None,
        model: Optional[Any] = None,
        callback: Optional[Callable[[Any], Any]] = None,
    ):

        self.name = name
        self._value = value

        self.model = model

        CallbackMixIn.__init__(self, [callback] if callback else [])

    def __repr__(self):
        return f"<{self.__class__.__name__!s} ({self.name}): {self.value}>"

    def __str__(self):
        return str(self.value)

    @property
    def value(self) -> Any:
        """The value associated to the key."""

        return self._value

    @value.setter
    def value(self, new_value: Any):
        """Sets the value of the key and schedules the callback."""

        self._value = new_value
        self.notify(self)

    def flatten(self) -> Dict[str, Any]:
        """Returns a dictionary with the name and value of the property."""

        return {self.name: self.value}


T = TypeVar("T", bound=Property)


class BaseModel(CaseInsensitiveDict[T], CallbackMixIn):
    """A JSON-compliant model.

    Parameters
    ----------
    name
        The name of the model.
    callback
        A function or coroutine to call when the datamodel changes. The
        function is called with the instance of `.BaseModel` and the key that
        changed.
    """

    def __init__(
        self,
        name: str,
        callback: Optional[Callable[[Any], Any]] = None,
        **kwargs,
    ):

        self.name = name

        CaseInsensitiveDict.__init__(self, {})
        CallbackMixIn.__init__(self, [callback] if callback else [])

    def __repr__(self):
        return f"<Model ({self.name})>"

    def __str__(self):
        return str(self.flatten())

    def flatten(self) -> Dict[str, Any]:
        """Returns a dictionary of values.

        Return a dictionary in which the `Property` instances are replaced
        with their values.
        """

        return {key: prop.value for key, prop in self.items()}

    def jsonify(self) -> str:
        """Returns a JSON string with the model."""

        return json.dumps(self.flatten())


class Model(BaseModel[Property]):
    """A model with JSON validation.

    In addition to the parameters in `.BaseModel`, the following parameters
    are accepted:

    Parameters
    ----------
    schema
        A valid JSON schema, to be used for validation.
    is_file
        Whether the input schema is a filepath or a dictionary.
    """

    VALIDATOR = jsonschema.Draft7Validator

    def __init__(self, name: str, schema: SchemaType, is_file: bool = False, **kwargs):

        if is_file:
            schema = cast(PathLike, schema)
            schema = open(pathlib.Path(schema).expanduser(), "r").read()

        if isinstance(schema, str):
            try:
                schema = json.loads(schema)
            except json.JSONDecodeError:
                raise ValueError("cannot parse input schema.")

        self.schema = cast("Dict[str, Any]", schema)

        if not self.check_schema(self.schema):
            raise ValueError(f"schema {name!r} is invalid.")

        type_checker = self.VALIDATOR.TYPE_CHECKER.redefine(
            "array", lambda checker, instance: isinstance(instance, (list, tuple))
        )
        self.VALIDATOR = jsonschema.validators.extend(
            self.VALIDATOR,
            type_checker=type_checker,
        )
        self.validator = self.VALIDATOR(self.schema)

        if (
            "type" not in self.schema
            or self.schema["type"] != "object"
            or "properties" not in self.schema
        ):
            raise ValueError("Schema must be of type object.")

        # All models must have these keys.
        for prop in DEFAULT_SCHEMA:
            if prop not in self.schema["properties"]:
                self.schema["properties"][prop] = DEFAULT_SCHEMA[prop]

        super().__init__(name, **kwargs)

        for name in self.schema["properties"]:
            self[name] = Property(name, model=self)

    @staticmethod
    def check_schema(schema: Dict[str, Any]) -> bool:
        """Checks whether a JSON schema is valid.

        Parameters
        ----------
        schema
            The schema to check as a dictionary.

        Returns
        -------
        result
            Returns `True` if the schema is a valid JSON schema, `False`
            otherwise.
        """

        try:
            Model.VALIDATOR.check_schema(schema)
            return True
        except jsonschema.SchemaError:
            return False

    def update_model(self, instance: Dict[str, Any]):
        """Validates a new instance and updates the model."""

        try:
            self.validator.validate(instance)
        except jsonschema.exceptions.ValidationError as err:
            return False, err

        for key, value in instance.items():
            if key in self:
                if isinstance(self[key].value, dict) and isinstance(value, dict):
                    # Copy previous value and update it but then assign it to
                    # force the callback in the property.
                    new_value = self[key].value.copy()
                    new_value.update(value)
                    self[key].value = new_value
                else:
                    self[key].value = value
                self.notify(self, self[key])

        return True, None


class ModelSet(dict):
    """A dictionary of `.Model` instances.

    Given a list of ``actors``, queries each of the actors to return their
    own schemas, which are then parsed and loaded as `.Model` instances.
    Since obtaining the schema require sending a command to the actor, that
    process happens when the coroutine `.load_schemas` is awaited, which
    should usually occur when the client is started.

    Parameters
    ----------
    client
        A client with a connection to the actors to monitor.
    actors
        A list of actor models whose schemas will be loaded.
    get_schema_command
        The command to send to the actor to get it to return its own schema.
    raise_exception
        Whether to raise an exception if any of the models cannot be loaded.
    kwargs
        Keyword arguments to be passed to `Model`.

    Example
    -------

        >>> model_set = ModelSet(client, actors=['sop', 'guider'])
        >>> model_set['sop']
        <Model (name='sop')>

    """

    def __init__(
        self,
        client: clu.base.BaseClient,
        actors: List[str] = [],
        get_schema_command: str = "get_schema",
        raise_exception: bool = True,
        **kwargs,
    ):

        dict.__init__(self, {})

        self.client = client
        self.actors = actors

        self.__raise_exception = raise_exception
        self.__get_schema_command = get_schema_command
        self.__kwargs = kwargs

    async def load_schemas(self, actors: Optional[List[str]] = None):
        """Loads the actor schames."""

        actors = actors or self.actors or []
        schema = None

        for actor in actors:

            try:

                cmd = await self.client.send_command(actor, self.__get_schema_command)
                await cmd

                if cmd.status.did_fail:
                    raise CluError(f"Failed getting schema for {actor}.")
                else:
                    for reply in cmd.replies:
                        if "schema" in reply.body:
                            schema = json.loads(reply.body["schema"])
                            break
                    if schema is None:
                        raise CluError(f"{actor} did not reply with a model.")

                self[actor] = Model(actor, schema, **self.__kwargs)

            except Exception as err:

                if not self.__raise_exception:
                    warnings.warn(f"Cannot load model {actor!r}. {err}", CluWarning)
                    continue
                raise



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2018-01-17
# @Filename: command.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import re
import time
from contextlib import suppress

from typing import Any, Callable, Dict, Generic, List, Optional, Tuple, TypeVar, Union

import clu
import clu.base
from clu.tools import CommandStatus, StatusMixIn


__all__ = [
    "BaseCommand",
    "Command",
    "parse_legacy_command",
    "TimedCommand",
    "TimedCommandList",
]


Actor_co = TypeVar("Actor_co", bound="clu.base.BaseActor", covariant=True)


class BaseCommand(asyncio.Future, StatusMixIn[CommandStatus], Generic[Actor_co]):
    """Base class for commands of all types (user and device).

    A `BaseCommand` instance is a `~asyncio.Future` whose result gets set
    when the status is done (either successfully or not).

    Parameters
    ----------
    commander_id
        The ID of the commander issuing this command. Can be a string or an
        integer. Normally the former is used for new-style actor and the latter
        for legacy actors.
    command_id
        The ID associated to this command. As with the commander_id, it can be
        a string or an integer
    consumer_id
        The actor that is consuming this command. Normally this is our own
        actor but if we are commanding another actor ``consumer_id`` will
        be the destination actor.
    actor
        The actor instance associated to this command.
    parent
        Another `.BaseCommand` object that is issuing this subcommand.
        Messages emitted by the command will use the parent ``command_id``.
    status_callback
        A function to call when the status changes.
    call_now
        Whether to call ``status_callback`` when initialising the command.
    default_keyword
        The keyword to use when writing a message that does not specify a
        keyword.
    loop
        The event loop.

    """

    def __init__(
        self,
        commander_id: Union[int, str] = 0,
        command_id: Union[int, str] = 0,
        consumer_id: Union[int, str] = 0,
        actor: Optional[Actor_co] = None,
        parent: Optional[BaseCommand[Actor_co]] = None,
        status_callback: Optional[Callable[[CommandStatus], Any]] = None,
        call_now: bool = False,
        default_keyword: str = "text",
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ):

        self.commander_id = commander_id
        self.consumer_id = consumer_id
        self.command_id = command_id

        self.actor = actor
        self.parent = parent

        self.default_keyword = default_keyword
        self.loop = loop or asyncio.get_event_loop()

        #: .Reply: A list of replies this command has received.
        self.replies: List[clu.base.Reply] = []

        asyncio.Future.__init__(self, loop=self.loop)

        self._status: CommandStatus
        StatusMixIn.__init__(
            self,
            CommandStatus,
            initial_status=CommandStatus.READY,
            callback_func=status_callback,
            call_now=call_now,
        )

    @property
    def status(self) -> Union[CommandStatus, None]:
        """Returns the status."""

        return self._status

    @status.setter
    def status(self, status: CommandStatus):
        """Sets the status. A message is output to the users.

        This setter calls `.set_status` with an empty message to the users.

        Parameters
        ----------
        status
            The status to set, either as a `CommandStatus` value or the
            integer associated with the maskbit. If ``value`` is a string,
            loops over the bits in `CommandStatus` and assigns the one whose
            name matches.
        """

        self.set_status(status)

    def set_status(
        self,
        status: Union[CommandStatus, str],
        message: Dict[str, Any] = None,
        silent: bool = False,
        **kwargs,
    ) -> BaseCommand:
        """Same as `.status` but allows to specify a message to the users."""

        if self.status.is_done:
            raise RuntimeError("cannot modify a done command.")

        if isinstance(status, str):
            for bit in self.flags:
                if status.lower() == bit.name.lower():
                    status = bit
                    break

        status = self.flags(status)

        try:
            is_flag = status in self.flags
            if not is_flag:
                raise TypeError()
        except TypeError:
            raise TypeError(f"Status {status!r} is not a valid command status.")

        if status != self._status:

            status_code = status.code
            if status_code is None:
                raise ValueError(f"Invalid status code {status_code!r}.")

            if self.actor and not silent:
                self.write(status_code, message, **kwargs)

            self._status = status

            self.do_callbacks()

            # If the command is done, set the result of the future.
            if self._status.is_done:
                self.set_result(self)

            # Set the status watcher
            if self.watcher is not None:
                self.watcher.set()

        return self

    def finish(self, *args, **kwargs):
        """Convenience method to mark a command `~.CommandStatus.DONE`."""

        self.set_status(CommandStatus.DONE, *args, **kwargs)
        return self

    def fail(self, *args, **kwargs):
        """Convenience method to mark a command `~.CommandStatus.FAILED`."""

        self.set_status(CommandStatus.FAILED, *args, **kwargs)
        return self

    def debug(self, *args, **kwargs):
        """Writes a debug-level message."""

        self.write("d", *args, **kwargs)

    def info(self, *args, **kwargs):
        """Writes an info-level message."""

        self.write("i", *args, **kwargs)

    def warning(self, *args, **kwargs):
        """Writes a warning-level message."""

        self.write("w", *args, **kwargs)

    def error(self, *args, **kwargs):
        """Writes an error-level message (does not fail the command)."""

        self.write("e", *args, **kwargs)

    def write(
        self,
        message_code: str = "i",
        message: Optional[Union[Dict[str, Any], str]] = None,
        broadcast: bool = False,
        **kwargs,
    ):
        """Writes to the user(s).

        Parameters
        ----------
        message_code
            The message code (e.g., ``'i'`` or ``':'``).
        message
            The text to be output. If `None`, only the code will be written.
        """

        if message is None:
            message = {}
        elif isinstance(message, dict):
            pass
        elif isinstance(message, str):
            message = {self.default_keyword: message}
        else:
            raise ValueError(f"invalid message {message!r}")

        if not self.actor:
            raise clu.CommandError(
                "An actor has not been defined for "
                "this command. Cannot write to users."
            )

        command = self if not self.parent else self.parent

        self.actor.write(
            message_code,
            message=message,
            command=command,
            broadcast=broadcast,
            **kwargs,
        )


class Command(BaseCommand[Actor_co]):
    """A command from a user.

    Parameters
    ----------
    command_string
        The string that defines the body of the command.
    transport
        The TCP transport associated with this command (only relevant
        for `.LegacyActor` commands).
    """

    def __init__(
        self,
        command_string: str = "",
        transport: Optional[Any] = None,
        **kwargs,
    ):

        BaseCommand.__init__(self, **kwargs)

        #: The raw command string.
        self.raw_command_string = command_string

        #: The body of the command.
        self.body = command_string

        if not self.actor and kwargs.get("parent", None):
            self.actor = self.parent.actor

        #: The `~click.Context` running this command. Only relevant if
        #: using the built-in click-based parser.
        self.ctx = None

        self.transport = transport

    def parse(self) -> Command:
        """Parses the command."""

        if not isinstance(self.actor, clu.base.BaseActor):
            raise clu.CluError("the actor is not defined. Cannot parse command.")

        self.actor.parse_command(self)

        return self

    def __str__(self):
        return (
            f"<Command (commander_id={self.commander_id!r}, "
            f"command_id={self.command_id!r}, body={self.body!r})>"
        )


def parse_legacy_command(command_string: str) -> Tuple[int, str]:
    """Parses a command received by a legacy actor.

    Parameters
    ----------
    command_string
        The command to parse, including an optional header.

    Returns
    -------
    command_id, command_body
        The command ID, and the command body parsed from the command
        string.

    """

    _HEADER_BODY_RE = re.compile(
        r"((?P<cmdID>\d+)(?:\s+\d+)?\s+)?((?P<cmdBody>[A-Za-z_].*))?$"
    )

    command_match = _HEADER_BODY_RE.match(command_string)
    if not command_match:
        raise clu.CommandError(f"Could not parse command {command_string!r}")

    command_dict = command_match.groupdict("")

    command_id_str = command_dict["cmdID"]
    if command_id_str:
        command_id = int(command_id_str)
    else:
        command_id = 0

    command_body = command_dict.get("cmdBody", "").strip()

    return command_id, command_body


class TimedCommandList(list):
    """A list of `.TimedCommand` objects that will be executed on a loop.

    Parameters
    ----------
    actor
        The actor in which the commands are to be run.
    resolution
        In seconds, how frequently to check if any of the `.TimedCommand` must
        be executed.

    """

    def __init__(
        self,
        actor: clu.base.BaseActor,
        resolution=0.5,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ):

        self.resolution = resolution
        self.actor = actor
        self.loop = loop or asyncio.get_event_loop()

        self._task: Optional[asyncio.Task] = None

        list.__init__(self, [])

    def add_command(self, command_string: str, **kwargs):
        """Adds a new `.TimedCommand`."""

        self.append(TimedCommand(command_string, **kwargs))

    async def poller(self):
        """The polling loop."""

        current_time = time.time()

        while True:

            for timed_command in self:
                elapsed = current_time - timed_command.last_run
                if elapsed > timed_command.delay:
                    timed_command_task = self.loop.create_task(
                        timed_command.run(self.actor)
                    )
                    timed_command_task.add_done_callback(timed_command.done)

            self._sleep_task = self.loop.create_task(asyncio.sleep(self.resolution))

            await self._sleep_task
            current_time += self.resolution

    def start(self) -> TimedCommandList:
        """Starts the loop."""

        if self.running:
            raise RuntimeError("poller is already running.")

        self._task = self.loop.create_task(self.poller())

        return self

    async def stop(self):
        """Cancel the poller."""

        if not self.running:
            return

        if self._task:
            self._task.cancel()
            with suppress(asyncio.CancelledError):
                await self._task

    @property
    def running(self) -> bool:
        """Returns `True` if the poller is running."""

        if self._task and not self._task.cancelled():
            return True

        return False


class TimedCommand(object):
    """A command to be executed on a loop.

    Parameters
    ----------
    command_string
        The command string to run.
    delay
        How many seconds to wait between repeated calls.
    """

    def __init__(self, command_string: str, delay: float = 1):

        self.command_string = command_string
        self.delay = delay

        self.last_run = 0.0

    async def run(self, actor: clu.base.BaseActor):
        """Run the command."""

        await Command(self.command_string, actor=actor, commander_id=actor.name).parse()

    def done(self, task):
        """Marks the execution of a command."""

        self.last_run = time.time()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-05-10
# @Filename: tron.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import logging

from typing import Any, Callable, List, Optional

from clu.base import BaseClient
from clu.command import Command, CommandStatus
from clu.model import BaseModel, Property
from clu.protocol import ReconnectingTCPClientProtocol

from .types.keys import Key, KeysDictionary
from .types.messages import Keyword
from .types.parser import ParseError, ReplyParser


__all__ = ["TronConnection", "TronModel", "TronKey"]


class TronKey(Property):
    """A Tron model key with callbacks.

    Similar to `.Property` but stores the original key with the keyword
    datamodel.
    """

    def __init__(
        self,
        name: str,
        key: Key,
        keyword: Optional[Keyword] = None,
        model: Optional[TronModel] = None,
        callback: Optional[Callable[[TronKey], Any]] = None,
    ):

        initial_value = [None] * len(key.typedValues.vtypes)
        super().__init__(name, value=initial_value, model=model, callback=callback)

        self.key = key
        self.keyword = None

        self.update_keyword(keyword)

    def update_keyword(self, keyword: Optional[Keyword]):
        """Updates the keyword and value."""

        if keyword is None:
            return

        self.keyword = keyword
        self.value = [value.native for value in keyword.values]

    def __getitem__(self, sl):
        return self.value.__getitem__(sl)


class TronModel(BaseModel[TronKey]):
    """A JSON-compliant model for actor keywords.

    Parameters
    ----------
    keydict
        A dictionary of keys that define the datamodel.
    callback
        A function or coroutine to call when the datamodel changes. The
        function is called with the instance of `.TronModel` and the modified keyword.
        If the callback is a coroutine, it is scheduled as a task.
    log
        Where to log messages.
    """

    def __init__(
        self,
        keydict: KeysDictionary,
        callback: Callable[[TronModel], Any] = None,
        log: Optional[logging.Logger] = None,
    ):

        super().__init__(keydict.name, callback=callback, log=log)

        self.keydict = keydict

        for key in self.keydict.keys:
            key = self.keydict.keys[key]
            self[key.name] = TronKey(key.name, key, model=self)

    def reload(self):
        """Reloads the model. Clears callbacks."""

        model = self.keydict.name
        keydict = KeysDictionary.load(model)

        self.__init__(keydict)

    def parse_reply(self, reply):
        """Parses a reply and updates the datamodel."""

        for reply_key in reply.keywords:

            key_name = reply_key.name.lower()
            if key_name not in self.keydict:
                raise ParseError(
                    f"Cannot parse unknown keyword {self.name}.{reply_key.name}."
                )

            # When parsed the values in reply_key are string. After consuming
            # it with the Key, the values become typed values.
            result = self.keydict.keys[key_name].consume(reply_key)

            if not result:
                raise ParseError(
                    f"Failed parsing keyword {self.name}.{reply_key.name}."
                )

            self[key_name].update_keyword(reply_key)

            self.notify(self, self[key_name])


class TronLoggingFilter(logging.Filter):
    """Logs issues with the Tron parser only to the file logger."""

    def filter(self, record):
        return not record.getMessage().startswith("Failed parsing reply")


class TronClientProtocol(ReconnectingTCPClientProtocol):
    """A reconnecting protocol for the Tron connection."""

    def __init__(self, on_received, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._on_received = on_received

    def data_received(self, data):
        self._loop.call_soon(self._on_received, data)


class TronConnection(BaseClient):
    """Allows to send commands to Tron and manages the feed of replies.

    Parameters
    ----------
    name
        The name of the client.
    host
        The host on which Tron is running.
    port
        The port on which Tron is running.
    models
        A list of strings with the actors whose models will be tracked.
    kwargs
        Arguments to be passed to `.BaseClient`.
    """

    def __init__(
        self,
        host: str,
        port: int = 6093,
        name: str = "tron",
        models: List[str] = [],
        **kwargs,
    ):

        super().__init__(name, **kwargs)

        self.host = host
        self.port = port

        self._mid = 1

        models = models or []

        #: dict: The `KeysDictionary` associated with each actor to track.
        self.keyword_dicts = {model: KeysDictionary.load(model) for model in models}

        #: dict: The model and values of each actor being tracked.
        self.models = {model: TronModel(self.keyword_dicts[model]) for model in models}

        self.rparser: Any = ReplyParser()

        self.transport: asyncio.Transport | None = None
        self.protocol: TronClientProtocol | None = None

        self.running_commands = {}

        self.buffer = b""

        # We want to log problems with the Tron parser, but not to the console.
        if self.log.sh:
            self.log.sh.addFilter(TronLoggingFilter())

    async def start(self, get_keys=True):
        """Starts the connection to Tron.

        Parameters
        ----------
        get_keys : bool
            If `True`, gets all the keys in the models.
        """

        loop = asyncio.get_running_loop()
        self.transport, self.protocol = await loop.create_connection(  # type: ignore
            lambda: TronClientProtocol(self._handle_reply),
            self.host,
            self.port,
        )

        if get_keys:
            asyncio.create_task(self.get_keys())

        return self

    def stop(self):
        """Closes the connection."""

        self.transport.close()

    def connected(self):
        """Checks whether the client is connected."""

        if self.transport is None:
            return False

        return not self.transport.is_closing()

    async def run_forever(self):  # pragma: no cover

        # Keep alive until the connection is closed.
        while True:
            await asyncio.sleep(1)
            if self.transport.is_closing():
                return

    def send_command(self, target, command_string, commander="tron.tron", mid=None):
        """Sends a command through the hub.

        Parameters
        ----------
        target
            The actor to command.
        command_string
            The command to send.
        commander
            The actor or client sending the command. The format for Tron is
            "commander message_id target command" where commander needs to
            start with a letter and have a program and a user joined by a dot.
            Otherwise the command will be accepted but the reply will fail
            to parse.
        mid
            The message id. If `None`, a sequentially increasing value will
            be used. You should not specify a ``mid`` unless you really know
            what you're doing.
        """

        mid = mid or self._mid

        # The mid must be a 32-bit unsigned number.
        if mid >= 2 ** 32:
            self._mid = mid = mid % 2 ** 32

        command_string = f"{commander} {mid} {target} {command_string}\n"

        command = Command(command_string=command_string)
        command.set_status("RUNNING")
        self.running_commands[mid] = command

        self.transport.write(command_string.encode())

        self._mid += 1

        return command

    async def get_keys(self):
        """Gets all the keys for the models being tracked."""

        # Number of keys to be requested at once
        n_keys = 10

        for model in self.models.values():

            actor = model.name
            keys = [key.lower() for key in model]

            for ii in range(0, len(keys), n_keys):

                keys_to_request = keys[ii : ii + n_keys]

                if len(keys_to_request) == 0:
                    break

                keys_joined = " ".join(keys_to_request)

                command_string = f"getFor={actor} {keys_joined}"

                self.send_command("keys", command_string)

    def _handle_reply(self, data: bytes):
        """Tracks new replies from Tron and updates the model."""

        self.buffer += data

        lines = self.buffer.splitlines()
        if not self.buffer.endswith(b"\n"):
            self.buffer = lines.pop()
        else:
            self.buffer = b""

        for line in lines:
            try:
                # Do not strip here or that will cause parsing problems.
                line = line.decode()
                reply = self.rparser.parse(line)
            except ParseError:
                self.log.warning(f"Failed parsing reply '{line.strip()}'.")
                continue

            actor = reply.header.actor

            # The keys command returns keywords as if from the actor
            # keys_<actor> (e.g. keys_tcc).
            if actor.startswith("keys_"):
                actor = actor.split("_")[1]

            if actor in self.models:
                try:
                    self.models[actor].parse_reply(reply)
                except ParseError as ee:
                    self.log.warning(
                        f"Failed parsing reply {reply!r} with error: {ee!s}"
                    )

            mid = reply.header.commandId
            status = CommandStatus.code_to_status(reply.header.code.lower())

            if mid in self.running_commands:
                self.running_commands[mid].replies.append(reply)
                self.running_commands[mid].set_status(status)
                if status.is_done:
                    self.running_commands.pop(mid)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2017-12-05
# @Filename: command.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)


__all__ = [
    "CluError",
    "CluNotImplemented",
    "CluWarning",
    "CluDeprecationWarning",
    "CommandError",
    "CommandParserError",
]


class CluError(Exception):
    """A custom core Clu exception"""

    def __init__(self, message=None):

        message = "There has been an error" if not message else message

        super(CluError, self).__init__(message)


class CluNotImplemented(CluError):
    """A custom exception for not yet implemented features."""

    def __init__(self, message=None):

        message = "This feature is not implemented yet." if not message else message

        super(CluNotImplemented, self).__init__(message)


class CommandError(CluError):
    """An error raised when a `Command` fails."""

    pass


class CommandParserError(CluError):
    """An error raised when parsing a command fails."""

    pass


class CluBaseWarning(Warning):
    """Base warning for Clu."""

    pass


class CluWarning(UserWarning, CluBaseWarning):
    """The primary warning class."""

    pass


class CluDeprecationWarning(CluBaseWarning):
    """A warning for deprecated features."""

    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-05-20
# @Filename: __main__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import asyncio
import datetime
import json
import os
import uuid

import aio_pika
import click
import prompt_toolkit
import prompt_toolkit.styles
import pygments
import pygments.lexers
import pygments.styles
from prompt_toolkit import print_formatted_text as print
from prompt_toolkit.formatted_text import PygmentsTokens
from prompt_toolkit.history import FileHistory
from prompt_toolkit.patch_stdout import patch_stdout

import clu


loop = asyncio.get_event_loop()


style = prompt_toolkit.styles.style_from_pygments_cls(
    pygments.styles.get_style_by_name("solarized-dark")
)


color_codes = {
    ">": "lightblue",
    "i": "lightblue",
    "d": "gray",
    "w": "yellow",
    "e": "LightCoral",
    "f": "red",
    ":": "green",
}


class ShellClient(clu.AMQPClient):
    """A shell client."""

    indent = False
    show_time = True
    ignore_broadcasts = False

    async def handle_reply(self, message: aio_pika.IncomingMessage):
        """Prints the formatted reply."""

        reply = await super().handle_reply(message)

        if reply is None:
            return

        commander_id = reply.info["headers"].get("commander_id", None)
        routing_key = message.routing_key
        is_broadcast = routing_key == "reply.broadcast" or reply.command_id is None

        if commander_id and commander_id != self.name and not is_broadcast:
            return

        if self.ignore_broadcasts and is_broadcast:
            return

        message_info = reply.info
        headers = message_info["headers"]

        message_code = headers.get("message_code", "")
        sender = headers.get("sender", "")

        message_code_esc = message_code if message_code != ">" else "&gt;"

        message_code_formatted = prompt_toolkit.formatted_text.HTML(
            f'<style font-weight="bold" '
            f'fg="{color_codes[message_code]}">{message_code_esc}</style>'
        )

        body = reply.body

        indent = 4 if self.indent is True else None
        body_tokens = list(
            pygments.lex(
                json.dumps(body, indent=indent),
                lexer=pygments.lexers.JsonLexer(),  # type: ignore
            )
        )

        if self.show_time:
            time = datetime.datetime.utcnow().isoformat().split("T")[1]
            time = time[0:12]  # Milliseconds
            print(
                prompt_toolkit.formatted_text.HTML(f'<style fg="Gray">{time}</style> '),
                end="",
            )

        if sender:
            print(f"{sender} ", end="")

        if message_code:
            print(message_code_formatted, end="")
            print(" ", end="")

        if body:
            print(PygmentsTokens(body_tokens), end="", style=style)
        else:
            print()  # Newline


async def shell_client_prompt(
    url=None,
    user=None,
    password=None,
    host=None,
    port=None,
    indent=True,
    show_time=True,
    ignore_broadcasts=False,
):

    # Give each client a unique name to ensure the queues are unique.
    uid = str(uuid.uuid4()).split("-")[0]

    client = await ShellClient(
        f"shell_client_{uid}",
        url=url,
        user=user,
        password=password,
        host=host,
        port=port,
        log=None,
    ).start()

    client.indent = indent
    client.show_time = show_time
    client.ignore_broadcasts = ignore_broadcasts

    history = FileHistory(os.path.expanduser("~/.clu_history"))

    session = prompt_toolkit.PromptSession("", history=history)

    while True:
        with patch_stdout():
            try:
                text = await session.prompt_async("")
            except KeyboardInterrupt:
                break
            except EOFError:
                break
            else:
                text = text.strip()
                if text.startswith("quit"):
                    break
                elif text == "":
                    continue
                else:
                    chunks = text.split()
                    if len(chunks) < 2:
                        print(f"Invalid command {text!r}")
                        continue

                    actor = chunks[0]
                    command_string = " ".join(chunks[1:])

                    await client.send_command(actor, command_string)


@click.command(name="clu")
@click.option("--url", type=str, help="AMQP RFC3986 formatted broker address.")
@click.option(
    "--user",
    "-U",
    type=str,
    show_default=True,
    default="guest",
    help="The AMQP username.",
)
@click.option(
    "--password",
    "-P",
    type=str,
    show_default=True,
    default="guest",
    help="The AMQP password.",
)
@click.option(
    "--host",
    "-H",
    type=str,
    show_default=True,
    default="localhost",
    help="The host running the AMQP server.",
)
@click.option(
    "--port",
    "-p",
    type=int,
    show_default=True,
    default=5672,
    help="The port on which the server is running",
)
@click.option(
    "--no-indent",
    "-n",
    is_flag=True,
    default=False,
    help="Do not indent the output JSONs.",
)
@click.option(
    "--ignore-broadcasts",
    "-b",
    is_flag=True,
    default=False,
    help="Only show replies to client commands.",
)
@click.option(
    "--no-time",
    "-t",
    is_flag=True,
    default=False,
    help="Do not show the message time.",
)
def clu_cli(url, user, password, host, port, no_indent, no_time, ignore_broadcasts):
    """Runs the AMQP command line interpreter."""

    shell_task = loop.create_task(
        shell_client_prompt(
            url=url,
            user=user,
            password=password,
            host=host,
            port=port,
            indent=not no_indent,
            show_time=not no_time,
            ignore_broadcasts=ignore_broadcasts,
        )
    )
    loop.run_until_complete(shell_task)


def main():
    clu_cli(obj={})


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'python'}

packages = \
['clu', 'clu.legacy', 'clu.legacy.types', 'clu.legacy.types.ply', 'clu.parsers']

package_data = \
{'': ['*']}

install_requires = \
['aio_pika>=6.4.1,<7.0.0',
 'aiormq>=3.2.0,<4.0.0',
 'click>=8.0,<9.0',
 'jsonschema>=3.0.1,<4.0.0',
 'prompt_toolkit>=3.0.6,<4.0.0',
 'sdsstools>=0.2.0']

entry_points = \
{'console_scripts': ['clu = clu.__main__:main']}

setup_kwargs = {
    'name': 'sdss-clu',
    'version': '1.0.3',
    'description': 'A new protocol for SDSS actors.',
    'long_description': "`CLU <https://tron.fandom.com/wiki/Clu>`__\n==========================================\n\n|py| |pypi| |black| |Build Status| |docs| |Coverage Status|\n\n\n`CLU <https://tron.fandom.com/wiki/Clu>`_ implements a new protocol for SDSS actor while providing support for legacy-style actor.\n\n\nFeatures\n--------\n\n- Asynchronous API based on `asyncio <https://docs.python.org/3/library/asyncio.html>`_.\n- New-style actor with message passing based on `AMQP <https://www.amqp.org/>`_ and `RabbitMQ <https://rabbitmq.com>`_.\n- Legacy-style actor for TCP socket communication via `tron <https://github.com/sdss/tron>`__.\n- Tools for device handling.\n- Messages are validated JSON strings.\n- `click <https://click.palletsprojects.com/en/7.x/>`__-enabled command parser.\n\n\nInstallation\n------------\n\n``CLU`` can be installed using ``pip`` as\n\n.. code-block:: console\n\n    pip install sdss-clu\n\nor from source\n\n.. code-block:: console\n\n    git clone https://github.com/sdss/clu\n    cd clu\n    pip install .\n\n\nDevelopment\n^^^^^^^^^^^\n\n``clu`` uses `poetry <http://poetry.eustace.io/>`__ for dependency management and packaging. To work with an editable install it's recommended that you setup ``poetry`` and install ``clu`` in a virtual environment by doing\n\n.. code-block:: console\n\n    poetry install\n\nPip does not support editable installs with PEP-517 yet. That means that running ``pip install -e .`` will fail because ``poetry`` doesn't use a ``setup.py`` file. As a workaround, you can use the ``create_setup.py`` file to generate a temporary ``setup.py`` file. To install ``clu`` in editable mode without ``poetry``, do\n\n.. code-block:: console\n\n    pip install --pre poetry\n    pip install -U setuptools\n    python create_setup.py\n    pip install -e .\n\n\nQuick start\n-----------\n\nCreating a new actor with ``CLU`` is easy. To instantiate and run an actor you can simply do\n\n.. code-block:: python\n\n    import asyncio\n    from clu import AMQPActor\n\n    async def main(loop):\n        actor = await Actor('my_actor').start()\n        await actor.run_forever()\n\n    asyncio.run(main(loop))\n\nNext, head to the `Getting started <https://clu.readthedocs.io/en/latest/getting-started.html>`__ section for more information about using actors. More examples are available `here <https://clu.readthedocs.io/en/latest/examples.html>`__.\n\n\nWhy a new messaging protocol for SDSS?\n--------------------------------------\n\nSay whatever you want about it, the `current SDSS message passing protocol <https://clu.readthedocs.io/en/latest/legacy.html>`_ based on ``Tron``, ``opscore``, and ``actorcore`` is stable and robust. So, why should we replace it? Here is a list of reasons:\n\n- It reinvents the wheel. Ok, in all honesty ``Tron`` and ``opscore`` were written when wheel were still not completely circular, but the truth is that nowadays there are more robust, standard, and better documented technologies out there for message passing.\n- We can remove the need for a central hub product by relying in open-source message brokers such as `RabbitMQ <https://rabbitmq.com>`__.\n- ``Tron`` and ``opscore`` are Python 2 and it's unclear the amount of effort that would be needed to convert them to Python 3.\n- While there is some documentation for ``Tron`` and ``opscore``, and the code is well written, it's also cumbersome and difficult to modify by people that didn't write it. It's ultimately non-maintainable.\n- The ``opsctore``/``actorkeys`` datamodel is custom-built and extremely difficult to maintain. Standard solutions such as JSON with a `JSON schema <https://json-schema.org/>`__ validator should be preferred.\n- `asyncio <https://docs.python.org/3/library/asyncio.html>`__ provides an asynchronous API that is cleaner and easier to code than using threads. It is also more readable and less convoluted than `twisted <https://twistedmatrix.com/trac/>`__ and it's a Python core library with very active development.\n- CLU uses `click <https://click.palletsprojects.com/en/7.x>`__ for parsing commands, providing a well-defined, easy to use parser.\n\n\n.. |Build Status| image:: https://img.shields.io/github/workflow/status/sdss/clu/Test\n    :alt: Build Status\n    :target: https://github.com/sdss/clu/actions\n\n.. |Coverage Status| image:: https://codecov.io/gh/sdss/clu/branch/main/graph/badge.svg\n    :alt: Coverage Status\n    :target: https://codecov.io/gh/sdss/clu\n\n.. |py| image:: https://img.shields.io/badge/python-3.7%20|%203.8%20|%203.9-blue\n    :alt: Python Versions\n    :target: https://docs.python.org/3/\n\n.. |docs| image:: https://readthedocs.org/projects/docs/badge/?version=latest\n    :alt: Documentation Status\n    :target: https://clu.readthedocs.io/en/latest/?badge=latest\n\n.. |pypi| image:: https://badge.fury.io/py/sdss-clu.svg\n    :alt: PyPI version\n    :target: https://badge.fury.io/py/sdss-clu\n\n.. |black| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n    :target: https://github.com/psf/black\n",
    'author': 'José Sánchez-Gallego',
    'author_email': 'gallegoj@uw.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sdss/clu',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-05-06
# @Filename: parser.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import functools
import inspect
import json
import re
import shlex

from typing import Any, List, TypeVar

import click
from click.decorators import group, pass_obj

from sdsstools.logger import SDSSLogger

from clu.command import Command

from .. import actor


__all__ = ["CluCommand", "CluGroup", "command_parser", "ClickParser", "timeout"]


def coroutine(fn):
    """Create a coroutine. Avoids deprecation of asyncio.coroutine in 3.10."""

    if inspect.iscoroutinefunction(fn):
        return fn

    @functools.wraps(fn)
    async def _wrapper(*args, **kwargs):
        return fn(*args, **kwargs)

    return _wrapper


class CluCommand(click.Command):
    """Override :py:class:`click.Command` to pass the actor and command."""

    def __init__(self, *args, context_settings=None, **kwargs):

        # Unless told otherwise, set ignore_unknown_options=True to prevent
        # negative numbers to be considered as options. See #40.
        if context_settings is None or "ignore_unknown_options" not in context_settings:
            context_settings = context_settings or {}
            context_settings["ignore_unknown_options"] = True

        super().__init__(
            *args,
            context_settings=context_settings,  # type: ignore
            **kwargs,
        )

    def done_callback(self, task, exception_handler=None):
        """Checks if the command task has been successfully done."""

        ctx = task.ctx

        command = ctx.obj["parser_args"][0]

        if exception_handler and task.exception():
            exception_handler(command, task.exception())

    async def _schedule_callback(self, ctx, timeout=None):
        """Schedules the callback as a task with a timeout."""

        parser_args = ctx.obj.get("parser_args", [])
        command = parser_args[0] if len(parser_args) > 0 else None

        callback_task = asyncio.create_task(
            ctx.invoke(self.callback, *parser_args, **ctx.params)
        )

        try:
            await asyncio.wait_for(callback_task, timeout=timeout)
        except asyncio.TimeoutError:
            if command:
                command.set_status(
                    command.status.TIMEDOUT,
                    f"command timed out after {timeout} seconds.",
                )
            return False

        return True

    def invoke(self, ctx):
        """As :py:class:`click.Command.invoke` but passes the actor and command."""

        if self.callback is not None:

            with ctx:

                loop = asyncio.get_event_loop()

                # Makes sure the callback is a coroutine
                if not asyncio.iscoroutinefunction(self.callback):
                    self.callback = coroutine(self.callback)

                # Check to see if there is a timeout value in the callback.
                # If so, schedules a task to be cancelled after timeout.
                timeout = getattr(self.callback, "timeout", None)

                log = ctx.obj.pop("log", None)

                # Defines the done callback function.
                exception_handler = ctx.obj.pop("exception_handler", None)
                done_callback = functools.partial(
                    self.done_callback, exception_handler=exception_handler
                )

                # Launches callback scheduler and adds the done callback
                ctx.task = loop.create_task(
                    self._schedule_callback(ctx, timeout=timeout)
                )
                ctx.task.add_done_callback(done_callback)

                # Add some attributes to the task because it's
                # what will be passed to done_callback
                ctx.task.ctx = ctx  # type: ignore
                ctx.task.log = log  # type: ignore

            return ctx


class CluGroup(click.Group):
    """Override :py:class:`click.Group`.

    Makes all child commands instances of `.CluCommand`.

    """

    def command(self, *args, **kwargs):
        """Override :py:class:`click.Group` to use `.CluCommand` by default."""

        if "cls" in kwargs:
            pass
        else:
            kwargs["cls"] = CluCommand

        def decorator(f):
            cmd = click.decorators.command(*args, **kwargs)(f)
            self.add_command(cmd)
            return cmd

        return decorator

    def parse_args(self, ctx, args):  # pragma: no cover

        # Copy this method so that we can turn off the printing of the
        # usage before ctx.exit()
        if not args and self.no_args_is_help and not ctx.resilient_parsing:
            ctx.exit()

        rest = click.Command.parse_args(self, ctx, args)
        if self.chain:
            ctx.protected_args = rest
            ctx.args = []
        elif rest:
            ctx.protected_args, ctx.args = rest[:1], rest[1:]

        return ctx.args

    def group(self, *args, **kwargs):
        """Creates a new group inheriting from this class."""

        if "cls" not in kwargs:
            kwargs["cls"] = self.__class__

        def decorator(f):
            assert not asyncio.iscoroutinefunction(f), "groups cannot be coroutines."
            cmd = group(*args, **kwargs)(f)
            self.add_command(cmd)
            return cmd

        return decorator


def timeout(seconds: float):
    """A decorator to timeout the command after a number of ``seconds``."""

    def decorator(f):

        # This is a bit of a hack but we cannot access the context here so
        # we add the timeout directly to the callback function.
        f.timeout = seconds

        async def helper(f, *args, **kwargs):
            if asyncio.iscoroutinefunction(f):
                return await f(*args, **kwargs)
            else:
                return f(*args, **kwargs)

        @functools.wraps(f)
        async def wrapper(*args, **kwargs):
            return await helper(f, *args, **kwargs)

        return functools.update_wrapper(wrapper, f)

    return decorator


def pass_args():
    """Thing wrapper around pass_obj to pass the command and parser_args."""

    def decorator(f):
        @functools.wraps(f)
        @pass_obj
        def new_func(obj, *args, **kwargs):
            return f(*obj["parser_args"], *args, **kwargs)

        return functools.update_wrapper(new_func, f)

    return decorator


@click.group(cls=CluGroup)
def command_parser(*args):
    pass


@command_parser.command()
def ping(*args):
    """Pings the actor."""

    command = args[0]
    command.set_status(command.status.DONE, "Pong.")

    return


@command_parser.command()
def version(*args):
    """Reports the version."""

    command = args[0]
    command.set_status(command.status.DONE, version=command.actor.version)

    return


@command_parser.command(cls=CluCommand, name="get_schema")
def get_schema(*args):
    """Returns the schema of the actor as a JSON schema."""

    command = args[0]

    if command.actor.model is None:
        return command.fail(error="The actor does not know its own schema.")

    return command.finish(schema=json.dumps(command.actor.model.schema))


@command_parser.command(name="help")
@click.argument("PARSER-COMMAND", type=str, required=False)
@click.pass_context
def help_(ctx, *args, parser_command):
    """Shows the help."""

    command = args[0]

    # The parser_command arrives wrapped in quotes to make sure is a single
    # value. Strip it and unpack it in as many groups and commands as needed.
    parser_command = parser_command.strip('"').split()

    help_lines = ""
    command_name = args[0].actor.name  # Actor name

    # Gets the help lines for the command group or for a specific command.
    if len(parser_command) > 0:

        ctx_commands = ctx.command.commands

        for ii in range(len(parser_command)):
            ctx_command_name = parser_command[ii].lower()
            command_name += f" {ctx_command_name}"
            if ctx_command_name not in ctx_commands:
                return command.fail(error=f"command {ctx_command_name} not found.")
            ctx_command = ctx_commands[ctx_command_name]
            if ii == len(parser_command) - 1:
                # This is the last element in the command list
                # so we want to actually output this help lines.
                help_lines = ctx_command.get_help(ctx)
            else:
                ctx_commands = ctx_command.commands

    else:

        help_lines: str = ctx.get_help()

    message = []
    for line in help_lines.splitlines():
        # Remove the parser name.
        match = re.match(r"^Usage: ([A-Za-z-_]+)", line)
        if match:
            line = line.replace(match.groups()[0], command_name)

        message.append(line)

    if isinstance(command.actor, (actor.AMQPActor, actor.JSONActor)):
        return command.finish(help=message)
    else:
        for line in message:
            command.warning(help=line)
        return command.finish()


@command_parser.command(name="keyword")
@click.argument("KEYWORD", type=str, required=True)
def keyword(command, *args, keyword):
    """Prints human-readable information about a keyword."""

    model = command.actor.model
    if model is None or model.schema is None:
        return command.fail(error="Actor does not have a data model.")

    if keyword not in model.schema["properties"]:
        return command.fail(error=f"Keyword {keyword!r} is not part of the data model.")

    schema = model.schema["properties"][keyword]

    lines = json.dumps(schema, indent=2).splitlines()[1:]
    max_length = max([len(line) for line in lines])

    command.info(text=f"{keyword} = {{".ljust(max_length, " "))
    for line in lines:
        command.info(text=line.replace('"', "").ljust(max_length, " "))

    command.finish()


T = TypeVar("T", bound=Command)


class ClickParser:
    """A command parser that uses Click at its base."""

    #: list: Arguments to be passed to each command in the parser.
    #: Note that the command is always passed first.
    parser_args: List[Any] = []
    parser = command_parser

    #: dict: Parameters to be set in the context object.
    context_obj = {}

    # For type hints
    log: SDSSLogger
    name: str

    def parse_command(self, command: T) -> T:
        """Parses a user command using the Click internals."""

        # This will pass the command as the first argument for each command.
        # If self.parser_args is defined, those arguments will be passed next.
        parser_args = [command]
        parser_args += self.parser_args

        # Empty command. Just finish the command.
        if not command.body:
            command.done()
            return command

        command.set_status(command.status.RUNNING)

        # If the command contains the --help flag,
        # redirects it to the help command.
        if "--help" in command.body:
            command.body = "help " + command.body
            command.body = command.body.replace(" --help", "")

        if not command.body.startswith("help"):
            command_args = shlex.split(command.body)
        else:
            command_args = ["help", '"{}"'.format(command.body[5:])]

        # We call the command with a custom context to get around
        # the default handling of exceptions in Click. This will force
        # exceptions to be raised instead of redirected to the stdout.
        # See http://click.palletsprojects.com/en/7.x/exceptions/
        obj = {
            "parser_args": parser_args,
            "log": self.log,
            "exception_handler": self._handle_command_exception,
        }
        obj.update(self.context_obj)
        ctx = self.parser.make_context(
            f"{self.name}-command-parser",
            command_args,
            obj=obj,
        )

        # Makes sure this is the global context. This solves problems when
        # the actor have been started from inside an existing context,
        # for example when it's called from a CLI click application.
        click.globals.push_context(ctx)

        try:
            self.parser.invoke(ctx)
        except Exception as exc:
            self._handle_command_exception(command, exc)

        return command

    def _handle_command_exception(self, command, exception):
        """Handles an exception during parsing or execution of a command."""

        try:

            raise exception

        except (click.ClickException) as ee:

            ctx = command.ctx
            message = ""

            # If this is a command that cannot be parsed.
            if not hasattr(ee, "message") and ctx:
                message = f"{ee.__class__.__name__}:\n{ctx.get_help()}"
            else:
                message = f"{ee.__class__.__name__}: {ee.format_message()}"

            if isinstance(command.actor, (actor.AMQPActor, actor.JSONActor)):
                command.warning(help=message.splitlines())
            else:
                lines = message.splitlines()
                for line in lines:
                    command.warning(help=line)

            msg = f"Command {command.body!r} failed."

            if not command.status.is_done:
                command.fail(error=msg)
            else:
                command.write("e", error=msg)

        except (click.exceptions.Exit, click.exceptions.Abort):

            if not command.status.is_done:
                command.fail(error=f"Command {command.body!r} was aborted.")

        except Exception as err:

            msg = (
                f"Command {command.body!r} failed because of an uncaught "
                f"error '{err.__class__.__name__}: {str(err)}'. See "
                f"traceback in the log for more information."
            )

            if command.status.is_done:
                command.write("i", text=msg)
            else:
                command.fail(error=msg)

            log = self.log or command.ctx.obj.get("log", None)
            if log:
                log.exception(f"Command {command.body!r} failed with error:")



"""
Message classes for SDSS-3 command and reply protocols

Refer to https://trac.sdss3.org/wiki/Ops/Messages for details.
"""

# type: ignore

# Created 10-Oct-2008 by David Kirkby (dkirkby@uci.edu)

import re

from . import types


class MessageError(Exception):
    pass


class Canonized(object):
    """
    Represents an object with a standard string representations

    The canonical string representation is used to define semantic
    equality via the __eq__ and __ne__ operators. The tokenized string
    representation defines grammatical equality.
    """

    def canonical(self):
        raise NotImplementedError

    def tokenized(self):
        raise NotImplementedError

    def __eq__(self, other):
        return isinstance(other, Canonized) and self.canonical() == other.canonical()

    def __ne__(self, other):
        return not isinstance(other, Canonized) or self.canonical() != other.canonical()


class Values(list, Canonized):
    """
    Represents the values associated with a command or keyword
    """

    # keyword or cmd values that do not match this pattern are quoted in canonical form
    unquoted = re.compile(r'[^"\'\s=,][^\s=,]*$')

    def canonical(self):
        """
        Returns the canonical string representation of a list of values

        Should not be used for the value associated with the "raw"
        keyword. A canonical value is quoted if it begins with a single
        or double quote or contains a comma, equals sign, or any inline
        white space. A single value is always quoted. A canonical string
        is, by definition, invariant under parsing.
        """
        result = ""
        for (comma, value) in enumerate(self):
            if comma:
                result += ","
            if isinstance(value, float):
                # use repr to capture the full precision
                value = repr(float(value))
            elif not isinstance(value, str):
                value = str(value)
            if self.unquoted.match(value) and (comma or len(self) > 1):
                # a lone value must be quoted so it is not confused with a keyword name
                result += value
            else:
                # only double quotes need to be escaped
                result += '"%s"' % value.decode("string_escape").replace('"', '\\"')
        return result

    def tokenized(self):
        """
        Returns the parse token representation of a list of values

        Should not be used for the value associated with the "raw"
        keyword. Does not distinguish between quoted and unquoted
        values. A tokenized string is invariant under parsing.
        """
        result = ""
        for (comma, value) in enumerate(self):
            if comma:
                result += ","
            result += "123"
        return result


class Keyword(Canonized):
    """
    Represents a keyword and its values
    """

    def __init__(self, name, values=None):
        """
        Creates a new keyword instance

        The "raw" keyword name (case insensitive) is reserved and must
        not be used. No checks are performed on the characters used in
        the name (the parser has normally already done this).
        """
        if name.lower() == "raw":
            raise MessageError('keyword "%s" is reserved' % name)
        self.name = name
        self.values = Values(values or [])
        self.matched = False

    def canonical(self):
        """
        Returns the canonical string representation of this keyword

        Two keywords with identical canonical strings are equivalent.
        The canonical form of a keyword name is all lower case. A
        canonical string is, by definition, invariant under parsing.
        """
        result = self.name.lower()
        if self.values and len(self.values) > 0:
            result += "=%s" % self.values.canonical()
        return result

    def tokenized(self):
        """
        Returns the parse token representation of a keyword

        Two keywords with identical tokenized strings have the same
        basic grammar. A tokenized string is invariant under parsing.
        """
        result = "KEY"
        if self.values and len(self.values) > 0:
            result += "=%s" % self.values.tokenized()
        return result

    def clone(self):
        """
        Returns a clone of this object
        """
        new = Keyword(self.name, self.values)
        new.matched = self.matched
        return new

    def copy(self, other):
        """
        Copies the attributes of another Keyword instance
        """
        self.name = other.name
        self.values = other.values
        self.matched = other.matched

    def __repr__(self):
        result = "KEY(%s)=%r" % (self.name, self.values)
        if self.matched:
            result = "{" + result + "}"
        return result

    def __getitem__(self, ii):
        """Returns the i-th value from the keyword."""

        return self.values[ii]

    # def __len__(self):
    #     """Returns the length of the values."""
    #
    #     return len(self.values)


class RawKeyword(Keyword):
    """
    Represents the reserved RAW keyword
    """

    def __init__(self, line):
        """
        Creates a new raw keyword instance
        """
        self.name = "raw"
        self.values = Values([line])
        self.matched = False

    def __repr__(self):
        result = "RAWKEY(%s)=%r" % (self.name, self.values)
        return result

    def canonical(self):
        """
        Returns the canonical string representation of a raw keyword
        """
        return "raw=%s" % self.values[0]

    def tokenized(self):
        """
        Returns the parse token representation of a raw keyword
        """
        return "RAW=LINE"

    def clone(self):
        """
        Returns a clone of this object
        """
        new = RawKeyword(self.values[0])
        new.matched = self.matched
        return new

    def copy(self, other):
        """
        Copies the attributes of another Keyword instance
        """
        self.values = other.values
        self.matched = other.matched


class Keywords(list, Canonized):
    """
    Represents an ordered set of keywords
    """

    def canonical(self, delimiter):
        """
        Returns the canonical string representation of a set of keywords

        Two sets of keywords with identical canonical strings are
        equivalent. Keyword ordering is considered to be significant. A
        canonical string is, by definition, invariant under parsing.
        """
        result = ""
        for delimited, keyword in enumerate(self):
            if delimited:
                result += delimiter
            result += keyword.canonical()
        return result

    def tokenized(self, delimiter):
        """
        Returns the parse token representation of a set of keywords

        Two sets of keywords with identical tokenized strings have the
        same basic grammar. Keyword ordering is considered to be
        significant. A tokenized string is invariant under parsing.
        """
        result = ""
        for delimited, keyword in enumerate(self):
            if delimited:
                result += delimiter
            result += keyword.tokenized()
        return result

    def clone(self):
        """
        Returns a clone of this object
        """
        new = Keywords()
        for keyword in self:
            new.append(keyword.clone())
        return new

    def copy(self, other):
        """
        Copies the keywords of another Keywords instance
        """
        del self[:]
        self.extend(other)

    def __getitem__(self, index):
        """
        Returns the keyword at the specified integer or string index

        An integer index returns a keyword by position starting from
        zero. A string index returns a keyword by name.
        """
        try:
            return list.__getitem__(self, index)
        except TypeError:
            if isinstance(index, str):
                name = index.lower()
                for k in self:
                    if k.name.lower() == name:
                        return k
                raise KeyError(index)
            raise TypeError("Keywords index must be integer or string")

    def get(self, name, default=None):
        """Returns the keyword matching name or default if not found."""

        for key in self:
            if key.name == name:
                return key

        return default

    def __getslice__(self, begin, end):
        """
        Returns a slice of ordered keywords via the usual list syntax
        """
        return Keywords(list.__getslice__(self, begin, end))

    def __contains__(self, name):
        """
        Tests if a keyword with the specified name is present
        """
        if isinstance(name, str):
            lower = name.lower()
            for k in self:
                if k.name.lower() == lower:
                    return True
            return False
        raise TypeError("Keyword names must be strings")


class ReplyHeader(Canonized):
    """
    Represents the headers of a reply
    """

    MsgCode = types.Enum(
        ">",
        "D",
        "I",
        "W",
        "E",
        ":",
        "F",
        "!",
        labelHelp=[
            "Queued",
            "Debug",
            "Information",
            "Warning",
            "Error",
            "Finished",
            "Error",
            "Fatal",
        ],
        name="code",
        help="Reply header status code",
    )

    def __init__(self, program, user, actorStack, commandId, actor, code):
        self.program = program
        self.user = user
        self.actorStack = actorStack
        # The commander name is the concatenation of the first three fields and should
        # exactly reproduce the initial word of the reply message string.
        self.cmdrName = "%s.%s%s" % (self.program, self.user, self.actorStack)
        self.commandId = int(commandId)
        self.actor = actor
        try:
            self.code = ReplyHeader.MsgCode(code)
        except ValueError:
            raise MessageError("Invalid reply header code: %s" % code)

    def canonical(self):
        return "%s %d %s %s" % (self.cmdrName, self.commandId, self.actor, self.code)

    def tokenized(self):
        return "prog.user 123 actor %s" % self.code

    def clone(self):
        return ReplyHeader(
            self.program, self.user, self.commandId, self.actor, self.code
        )

    def copy(self, other):
        self.program = other.program
        self.user = other.user
        self.commandId = other.commandId
        self.actor = other.actor
        self.code = other.code

    def __repr__(self):
        return "HDR(%s,%s,%d,%s,%s)" % (
            self.program,
            self.user,
            self.commandId,
            self.actor,
            self.code,
        )


class ActorReplyHeader(Canonized):
    """
    Represents the header of a reply from an actor (as opposed to the hub)
    """

    MsgCode = types.Enum(
        ">",
        "D",
        "I",
        "W",
        "E",
        ":",
        "F",
        "!",
        labelHelp=[
            "Queued",
            "Debug",
            "Information",
            "Warning",
            "Error",
            "Finished",
            "Error",
            "Fatal",
        ],
        name="code",
        help="Reply header status code",
    )

    def __init__(self, commandId, userId, code):
        self.commandId = int(commandId)
        self.userId = int(userId)
        try:
            self.code = ReplyHeader.MsgCode(code)
        except ValueError:
            raise MessageError("Invalid reply header code: %s" % code)

    def canonical(self):
        return "%d %d %s" % (self.commandId, self.userId, self.code)

    def tokenized(self):
        return "123 1 %s" % self.code

    def clone(self):
        return ReplyHeader(self.commandId, self.userId, self.code)

    def copy(self, other):
        self.commandId = other.commandId
        self.userId = other.userId
        self.code = other.code

    def __repr__(self):
        return "ACTHDR(%d,%d,%s)" % (self.commandId, self.userId, self.code)


class Reply(Canonized):
    """
    Represents a reply

    A reply consists of one or more keywords separated by semicolons.
    Keywords are ordered and duplicates do not generate a runtime
    exception.
    """

    def __init__(self, header, keywords, string=None):
        """
        Creates a new reply instance
        """
        self.header = header
        self.keywords = Keywords(keywords)
        self.string = string

    def canonical(self):
        """
        Returns the canonical string representation of this reply

        Two replies with identical canonical strings are equivalent. A
        canonical string is, by definition, invariant under parsing.
        """
        return "%s %s" % (
            self.header.canonical(),
            self.keywords.canonical(delimiter=";"),
        )

    def tokenized(self):
        """
        Returns the parse token representation of a reply

        Two replies with identical tokenized strings have the same
        basic grammar. A tokenized string is invariant under parsing.
        """
        return "%s %s" % (
            self.header.tokenized(),
            self.keywords.tokenized(delimiter=";"),
        )

    def clone(self):
        """
        Returns a clone of this object
        """
        new = Reply(header=self.header.clone(), keywords=[], string=self.string)
        new.keywords = self.keywords.clone()
        return new

    def copy(self, other):
        """
        Copies the keywords of another Keywords instance
        """
        self.header.copy(other.header)
        self.keywords = other.keywords
        self.string = other.string

    def __repr__(self):
        result = "REPLY(%r,%r)" % (self.header, self.keywords)
        return result


class CommandHeader(Canonized):
    """
    Represents the headers of a command
    """

    def __init__(self, cmdrName, mid, actor):
        self.cmdrName = cmdrName
        self.mid = mid
        self.actor = actor

    @property
    def program(self):
        """ Return program part of cmdrname. """
        return self.cmdrName.split(".", 1)[0]

    def canonical(self):
        return "%s %d %s" % (self.cmdrName, self.mid, self.actor)

    def tokenized(self):
        return "prog.user 123 actor %s" % self.actor

    def clone(self):
        return CommandHeader(self.cmdrName, self.mid, self.actor)

    def copy(self, other):
        self.cmdrName = other.cmdrName
        self.mid = other.mid
        self.actor = other.actor

    def __repr__(self):
        return "CommandHeader(%s,%d,%s)" % (self.cmdrName, self.mid, self.actor)


class Command(Canonized):
    """Represents a command

    A command consists of a name (verb) with optional values followed by
    zero or more keywords. Keywords are ordered and duplicates do not
    generate a runtime exception.

    """

    def __init__(self, name, values=None, keywords=None, string=None):
        """Creates a new command instance

        The name must not contain the dot (.) character and must not
        equal "raw" (case insensitive), but no other checks are
        performed on the characters used in the name (the parser has
        normally already done this).
        """
        if "." in name:
            raise MessageError("name %r contains an illegal dot (.) character." % name)
        if name.lower() == "raw":
            raise MessageError('name cannot be "%s"' % name)
        self.name = name
        self.values = Values(values or [])
        self.keywords = Keywords(keywords or [])
        self.string = string

    def canonical(self):
        """
        Returns the canonical string representation of this command

        Two commands with identical canonical strings are equivalent.
        The canonical form of a verb name is all lower case. A canonical
        string is, by definition, invariant under parsing.
        """
        result = self.name.lower()
        if self.values and len(self.values) > 0:
            result += " %s" % self.values.canonical()
        if self.keywords:
            result += " "
            result += self.keywords.canonical(delimiter=" ")
        return result

    def tokenized(self):
        """
        Returns the parse token representation of a command

        Two commands with identical tokenized strings have the same
        basic grammar. A tokenized string is invariant under parsing.
        """
        result = "VERB"
        if self.values and len(self.values) > 0:
            result += " %s" % self.values.tokenized()
        if self.keywords:
            result += " "
            result += self.keywords.tokenized(delimiter=" ")
        return result

    def clone(self):
        """
        Returns a clone of this object
        """
        new = Command(self.name, self.values, [], self.string)
        new.keywords = self.keywords.clone()
        return new

    def copy(self, other):
        """
        Copies the keywords of another Keywords instance
        """
        self.name = other.name
        self.values = other.values
        self.keywords = other.keywords
        self.string = other.string

    def __repr__(self):
        result = "CMD(%r=%r;%r)" % (self.name, self.values, self.keywords)
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2018-01-16
# @Filename: actor.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import json
import pathlib
import re
import uuid

from typing import Any, Dict, Optional, TypeVar, Union, cast

import aio_pika as apika
import click

from .base import BaseActor, Reply
from .client import AMQPClient
from .command import Command, TimedCommandList
from .exceptions import CommandError
from .parsers import ClickParser, CluCommand
from .protocol import TCPStreamServer
from .tools import log_reply


__all__ = ["AMQPActor", "JSONActor", "AMQPBaseActor", "TCPBaseActor"]


T = TypeVar("T")
PathLike = Union[str, pathlib.Path]
SchemaType = Union[Dict[str, Any], PathLike]


class CustomTransportType(asyncio.Transport):
    user_id: Union[str, int]
    multiline: bool


class AMQPBaseActor(AMQPClient, BaseActor):
    """An actor class that uses AMQP message brokering.

    This class differs from `~clu.legacy.actor.LegacyActor` in that it uses
    an AMQP messaging broker (typically RabbitMQ) to communicate with other
    actors in the system, instead of standard TCP sockets. Although the
    internals and protocols are different the entry points and behaviour for
    both classes should be almost identical.

    This class needs to be subclassed with a command parser.

    See the documentation for `.AMQPActor` and `.AMQPClient` for additional
    parameter information.

    Parameters
    ----------
    schema
        The path to the datamodel schema for the actor, in JSON Schema format.
        If the schema is provided all replies will be validated against it.
        An invalid reply will fail and not be emitted. The schema can also be
        set when subclassing by setting the class ``schema`` attribute.
    """

    def __init__(self, *args, schema: Optional[SchemaType] = None, **kwargs):

        AMQPClient.__init__(self, *args, **kwargs)

        self.commands_queue = None

        self.timed_commands = TimedCommandList(self)

        # Not calling BaseClient.__init__() here because we already called
        # AMQPClient.__init__.
        self.load_schema(schema)

    async def start(self, **kwargs):
        """Starts the connection to the AMQP broker."""

        # This sets the replies queue but not a commands one.
        await AMQPClient.start(self, **kwargs)

        # Binds the commands queue.
        self.commands_queue = await self.connection.add_queue(
            f"{self.name}_commands",
            callback=self.new_command,
            bindings=[f"command.{self.name}.#"],
        )

        self.log.info(
            f"commands queue {self.commands_queue.name!r} "
            f"bound to {self.connection.connection.url!s}"
        )

        self.timed_commands.start()

        return self

    async def new_command(self, message: apika.IncomingMessage, ack=True):
        """Handles a new command received by the actor."""

        if ack:
            async with message.process():
                headers = message.info()["headers"]
                command_body = json.loads(message.body.decode())
        else:
            headers = message.info()["headers"]
            command_body = json.loads(message.body.decode())

        commander_id = headers["commander_id"].decode()
        command_id = headers["command_id"].decode()
        command_string = command_body["command_string"]

        try:
            command = Command(
                command_string,
                command_id=command_id,
                commander_id=commander_id,
                consumer_id=self.name,
                actor=self,
                loop=self.loop,
            )
            command.actor = self  # Assign the actor
        except CommandError as ee:
            self.write(
                "f",
                {
                    "error": "Could not parse the "
                    "following as a command: "
                    f"{command_string!r}. {ee!r}"
                },
            )
            return

        return self.parse_command(command)

    async def _write_internal(self, reply: Reply):
        """Writes a message to user(s)."""

        message = reply.message
        message_json = json.dumps(message)

        command = reply.command

        if command is None or reply.broadcast:
            routing_key = "reply.broadcast"
        else:
            routing_key = f"reply.{command.commander_id}"

        commander_id = command.commander_id if command else None
        command_id = command.command_id if command else None

        headers = {
            "message_code": reply.message_code,
            "commander_id": commander_id,
            "command_id": command_id,
            "sender": self.name,
        }

        await self.connection.exchange.publish(
            apika.Message(
                message_json.encode(),
                content_type="text/json",
                headers=headers,
                correlation_id=command_id,
            ),
            routing_key=routing_key,
        )

        if self.log:
            log_reply(self.log, reply.message_code, message_json)


class AMQPActor(ClickParser, AMQPBaseActor):
    """An `AMQP actor <.AMQPBaseActor>` that uses a `click parser <.ClickParser>`."""

    pass


TCPBaseActor_co = TypeVar("TCPBaseActor_co", bound="TCPBaseActor")


class TCPBaseActor(BaseActor):
    """A TCP base actor that replies using JSON.

    This implementation of `.BaseActor` uses TCP as command/reply channel and
    replies to the user by sending a JSON-valid string. This makes it useful
    as a "device" actor that is not connected to the central message parsing
    system but that we still want to accept commands and reply with easily
    parseable messages.

    Commands received by this actor must be in the format
    ``[<uid>] <command string>``, where ``<uid>`` is any integer unique
    identifier that will be used as ``command_id`` and appended to any reply.

    This is a base actor that does not include a parser. It must be subclassed with
    a concrete parser that overrides ``parse_command``.

    Parameters
    ----------
    name
        The actor name.
    host
        The host where the TCP server will run.
    port
        The port of the TCP server.
    args,kwargs
        Arguments to be passed to `.BaseActor`.

    """

    def __init__(
        self,
        name: str,
        host: Optional[str] = None,
        port: Optional[int] = None,
        *args,
        **kwargs,
    ):

        super().__init__(name, *args, **kwargs)

        self.host = host or "localhost"
        self.port = port

        if self.port is None:
            raise ValueError("port needs to be specified.")

        #: Mapping of commander_id to transport
        self.transports = dict()

        #: TCPStreamServer: The server to talk to this actor.
        self.server = TCPStreamServer(
            self.host,
            self.port,
            loop=self.loop,
            connection_callback=self.new_user,
            data_received_callback=self.new_command,
        )

        self.timed_commands = TimedCommandList(self)

    async def start(self: TCPBaseActor_co) -> TCPBaseActor_co:
        """Starts the TCP server."""

        await self.server.start()
        self.log.info(f"running TCP server on {self.host}:{self.port}")

        self.timed_commands.start()

        return self

    async def stop(self):
        """Stops the client connection and running tasks."""

        if self.server.is_serving():
            self.server.stop()

        await self.timed_commands.stop()

    async def run_forever(self):
        """Runs the actor forever, keeping the loop alive."""

        await self.server.serve_forever()

    def new_user(self, transport: CustomTransportType):
        """Assigns userID to new client connection."""

        if transport.is_closing():
            if hasattr(transport, "user_id"):
                self.log.debug(f"user {transport.user_id} disconnected.")
                return self.transports.pop(transport.user_id)

        user_id = str(uuid.uuid4())
        transport.user_id = user_id
        transport.multiline = False
        self.transports[user_id] = transport

        sock = transport.get_extra_info("socket")
        if sock is not None:
            peername = sock.getpeername()[0]
            self.log.debug(f"user {user_id} connected from {peername}.")

    def new_command(self, transport: CustomTransportType, command_str: bytes):
        """Handles a new command received by the actor."""

        commander_id: Optional[int] = getattr(transport, "user_id", None)
        message: str = command_str.decode().strip()

        if not message:
            return

        command_id, command_string = re.match(r"([0-9]*)\s*(.+)", message).groups()

        if command_id == "":
            command_id = 0
        else:
            command_id = int(command_id)

        command_string = command_string.strip()

        if not command_string:
            return
        try:
            command = Command(
                command_string=command_string,
                commander_id=commander_id,
                command_id=command_id,
                consumer_id=self.name,
                actor=self,
                loop=self.loop,
                transport=transport,
            )
        except CommandError as ee:
            self.write(
                "f", {"text": f"Could not parse the following as a command: {ee!r}"}
            )
            return

        return self.parse_command(command)

    def write(self, *args, **kwargs):
        """Writes a message to user(s) as a JSON.

        A ``header`` keyword with the ``commander_id`` (i.e., the user id of
        the transport that sent the command), ``command_id``, ``message_code``,
        and ``sender`` is added to each message. The payload of the message
        is written to ``data``. An example of a valid message is

        .. code-block:: yaml

            {
                "header": {
                    "command_id": 0,
                    "commander_id": 1,
                    "message_code": "i",
                    "sender": "test_camera"
                },
                "message": {
                    "camera": {
                        "name": "test_camera",
                        "uid": "DEV_12345"
                    },
                    "status": {
                        "temperature": 25.0,
                        "cooler": 10.0
                    }
                }
            }

        Although the messsage is displayed here in multiple lines, it is
        written as a single line to the TCP clients to facilitate parsing.
        For a multiline output, which is more human-readable, use the
        ``multiline`` command.

        See `~.BaseActor.write` for details on the allowed parameters.
        """

        BaseActor.write(self, *args, **kwargs)

    def _write_internal(self, reply: Reply):
        """Write a reply to the users."""

        def send_to_transport(transport, message):

            if getattr(transport, "multiline", False):
                message_json = json.dumps(message, sort_keys=False, indent=4) + "\n"
            else:
                message_json = json.dumps(message, sort_keys=False) + "\n"
            transport.write(message_json.encode())

        message = reply.message
        command = cast(Command, reply.command)

        commander_id = command.commander_id if command else None
        command_id = command.command_id if command else None
        transport = command.transport if command else None

        message_full = {}
        header = {
            "header": {
                "command_id": command_id,
                "commander_id": commander_id,
                "message_code": reply.message_code,
                "sender": self.name,
            }
        }

        message_full.update(header)
        message_full.update({"data": message})

        if reply.broadcast or commander_id is None or transport is None:
            for transport in self.transports.values():
                send_to_transport(transport, message_full)
        else:
            send_to_transport(transport, message_full)

        message_json = json.dumps(message_full, sort_keys=False) + "\n"

        if self.log:
            log_reply(self.log, reply.message_code, message_json.strip())


class JSONActor(ClickParser, TCPBaseActor):
    """An implementation of `.TCPBaseActor` that uses a Click command parser."""

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        # Add the multiline command
        self.parser.add_command(multiline)

    def send_command(self, *args, **kwargs):
        """Not implemented for `.JSONActor`."""

        raise NotImplementedError("JSONActor cannot send commands to other actors.")


@click.command(cls=CluCommand)
@click.option("--on/--off", default=True, help="Turn multiline on/off.")
async def multiline(command: Command, on: bool):
    """Set multiline mode for the transport."""

    transport: CustomTransportType = getattr(command, "transport", None)
    if not transport:
        return command.fail("The command has no transport.")

    transport.multiline = on

    return command.finish("Multiline mode is {}".format("on" if on else "off"))
